//
// Created by sylwe on 21/09/2024.
//

#include "BiconnectedFinder.h"

#include "Solver.h"

Mask BiconnectedFinder::getNodesAfterBiconnectedReduction2(VVI &underV, Mask mask, int beg, int vt, VI path) {
    if( path.empty() ) path = Solver::findShortestPath(underV, beg, vt, mask);
    if( path.empty() ) return Mask(N);

    fill( ALL(was), false );
    fill( ALL(was2), false );
    fill( ALL(was_path), false );
    res.reset();
    for(int d : path) was_path[d] = true;
    mask.set(beg,false);

    tin[beg] = -1;

    function<void(int,int)> findBicomp = [&]( int u, int forbid ){
        bicomp.push_back(u);
        was2[u] = true;
        if(u != forbid) {
            for( int d : underV[u] ) if( !mask[d] && low[d] < tin[u] && !was2[d] ) findBicomp(d,forbid);
        }
    };

    int time_stamp = -1;
    function<void(int,int)> dfs = [&]( int u, int par ){
        was[u] = true;
        time_stamp++;
        tin[u] = time_stamp;
        low[u] = tin[u];

        for( int d : underV[u] ){
            if( !mask[d] && !was[d] ){
                dfs(d,u);
                if( low[d] == tin[u] ){
                    bicomp.clear();
                    findBicomp(d,u);
                    for(auto x : bicomp) was2[x] = false;

                    int steps = 0;
                    for( int x : bicomp ) if( was_path[x] ) steps++;
                    if( steps >= 2 ) for( int x : bicomp ) res.set(x,true);
                    else if( steps == 1 ) {
                        for( int x : bicomp ) if( was_path[x] ) res.set(x,true);
                    }
                }
            }
        }
        for(int d : underV[u]) if( d != par && !mask[d] ) low[u] = min(low[u], low[d]);
    };

    for( int i=0; i<N; i++ ) if( !mask[i] && !was[i] ) dfs(i,i);
    for( int i=0; i<N; i++ ) if( was_path[i]) res.set(i,true);
    if( beg != vt ) res.set(beg,false);
    return res;
}

Mask BiconnectedFinder::getNodesAfterBiconnectedReduction2FBS(VVI &underV, Mask mask, int beg, int vt, VI path,
    VI & ind_in_S, int v, VVB & ends) {

    if( path.empty() ) path = Solver::findShortestPath(underV, beg, vt, mask);
    if( path.empty() ) return Mask(N);

    fill( ALL(was), false );
    fill( ALL(was2), false );
    fill( ALL(was_path), false );
    res.reset();
    for(int d : path) was_path[d] = true;
    mask.set(beg,false);

    tin[beg] = -1;

    function<void(int,int)> findBicomp = [&]( int u, int forbid ){
        bicomp.push_back(u);
        was2[u] = true;
        if(u != forbid) {
            for( int d : underV[u] ) {
                if( ind_in_S[u] != -1 && ind_in_S[d] != -1 && u != v && d != v && !ends[u][d] ) continue;
                if( !mask[d] && low[d] < tin[u] && !was2[d] ) findBicomp(d,forbid);
            }
        }
    };

    int time_stamp = -1;
    function<void(int,int)> dfs = [&]( int u, int par ){
        was[u] = true;
        time_stamp++;
        tin[u] = time_stamp;
        low[u] = tin[u];

        for( int d : underV[u] ){
            if( !mask[d] && !was[d] ){
                if( ind_in_S[u] != -1 && ind_in_S[d] != -1 && u != v && d != v && !ends[u][d] ) continue;

                dfs(d,u);
                if( low[d] == tin[u] ){
                    bicomp.clear();
                    findBicomp(d,u);
                    for(auto x : bicomp) was2[x] = false;

                    int steps = 0;
                    for( int x : bicomp ) if( was_path[x] ) steps++;
                    if( steps >= 2 ) for( int x : bicomp ) res.set(x,true);
                    else if( steps == 1 ) {
                        for( int x : bicomp ) if( was_path[x] ) res.set(x,true);
                    }
                }
            }
        }
        for(int d : underV[u]) {
            if( ind_in_S[u] != -1 && ind_in_S[d] != -1 && u != v && d != v && !ends[u][d] ) continue;
            if( d != par && !mask[d] ) low[u] = min(low[u], low[d]);
        }
    };

    for( int i=0; i<N; i++ ) if( !mask[i] && !was[i] ) dfs(i,i);
    for( int i=0; i<N; i++ ) if( was_path[i]) res.set(i,true);
    if( beg != vt ) res.set(beg,false);
    return res;
}
