//
// Created by sylwe on 12.07.2024.
//

#include "Solver.h"

#include <ConnectedComponents.h>
#include <GraphInducer.h>
#include <GraphUtils.h>
#include <StandardUtils.h>

// #include "CCH.h"
#include "Kernelizer.h"
#include "CollectionOperators.h"
#include "Solver2.h"

Solver::Solver( VVI _V, int _L, int _vs, int _vt ) {
    V = _V;
    if(!V.empty()) {
        revV = GraphUtils::reverseGraph(V);
        N = V.size();
        inputL = L = min( _L, N-1 );
        vs = _vs, vt = _vt;
        dp.resize(N);
        dp_rev.resize(N);
    }
}

// template<bool use_caching, bool use_cch, bool use_sp_dp, bool use_state_merging>
// VBIG Solver::solveBTOptions() {
//     constexpr bool debug = true;
//
//     VBIG res(L+1,0);
//     VB was(N);
//     VI path;
//     backtracking_states_visited = 0;
//
//     constexpr int HS = (1<<10);
//     int HSmask = HS-1;
//     vector<map<HashType,Lengths>> cache(HS);
//     HashType cur_hash = 0;
//     ULL cache_hits = 0;
//
//     VVI  W(N);
//     for(int i=0; i<N; i++) W[i] = VI( V[i].size(),1 );
//     CCH cch(V,W);
//     if constexpr (use_cch) cch.constructNodeOrderAndBuild();
//
//     const auto zero_lengths = Lengths( VBIG(1,0),0 );
//
//     function<Lengths(int,int)> dfs = [&]( int num, int l ) {
//         backtracking_states_visited++;
//         if( l > L ) return zero_lengths;
//         if constexpr (use_cch) {
//             auto d = cch.dst(num,vt);
//             if( l + d > L ) return zero_lengths;
//             if constexpr (use_sp_dp) if( l + d == L ) {
//                 auto r = countShortestPaths( V, num, was );
//                 return Lengths( VBIG(1,r[vt]), d );
//             }
//         }
//
//         if(num == vt) return Lengths( VBIG(1,1),0 );
//
//         HashType state_hash = 0;
//         if constexpr (use_caching) {
//             cur_hash ^= node_hashes[num];
//             state_hash = cur_hash ^ node_hashes2[num];
//
//             auto it = cache[state_hash.hash & HSmask].find(cur_hash);
//             if( it != cache[state_hash.hash & HSmask].end()) {
//                 cache_hits++;
//                 cur_hash ^= node_hashes[num];
//                 return it->second;
//             }
//         }
//
//
//         if constexpr (debug) path.push_back(num);
//         was[num] = true;
//         if constexpr (use_cch) cch.excludeNode(num,was);
//
//         Lengths R( VBIG(1,0),0 );;
//         for( int d : V[num] ) {
//             if(!was[d]) {
//                 auto temp = dfs(d,l+1);
//                 R += temp;
//             }
//         }
//         was[num] = false;
//
//         if constexpr (use_caching) {
//             cache[state_hash.hash & HSmask][state_hash] = R;
//             cur_hash ^= node_hashes[num];
//         }
//
//         if constexpr (debug) path.pop_back();
//         if constexpr (use_cch) cch.includeNode(num,was);
//
//         return R;
//     };
//
//     auto lengths = dfs(vs,0);
//     for( int i=lengths.minl(); i<min(lengths.maxl()+1, (int)res.size()); i++ ) res[i] += lengths[i];
//
//     if constexpr (use_caching) {
//         int cache_size = accumulate(ALL(cache), 0, [](int s, auto & v){return s + v.size();});
//         DEBUG(cache_size);
//         DEBUG(cache_hits);
//     }
//     DEBUG(backtracking_states_visited);
//
//     return res;
// }

auto Solver::getDPRange( auto & dpa, auto val, auto fun ) {
    auto it1 = upper_bound( ALL(dpa), val, [&]( auto l, auto & sp )
                { return l< fun(sp) + 1; } );
    auto it2 = upper_bound( ALL(dpa), val, [&]( auto l, auto & sp )
        { return l < fun(sp); } );

    return make_pair(it1,it2);
}


VBIG Solver::solveBT() {
    VBIG res(L+1,0);
    VB was(N);
    // VI path;
    backtracking_states_visited = 0;

    function<void(int,int)> dfs = [&]( int num, int l ) {
        backtracking_states_visited++;
        if( l > L ) return;
        if(num == vt) { res[l]++;  return; }

        // path.push_back(num);
        // DEBUG(path);
        // DEBUG(V[num]);
        was[num] = true;
        for( int d : V[num] ) if(!was[d]) dfs(d,l+1);
        was[num] = false;

        // path.pop_back();
    };

    dfs(vs,0);

    return res;
}

void Solver::solveBruteBT( int num, VB & was, int l, bigint & res ) {
    backtracking_states_visited++;

    if( l > L ) return;
    // for(int i=0; i<l; i++) clog << "\t"; clog << num << endl;
    if(num == vt) {
        // clog << "FOUND" << endl;
        res++;
        return;
    }

    was[num] = true;
    for( int d : V[num] ) if(!was[d]) solveBruteBT(d,was,l+1, res);
    was[num] = false;
}

// finds shortest path from beg to end in graph G[ V \ S ]
VI Solver::findShortestPath(VVI & V, int beg, int end, Mask S) {
    auto [dst,par] = findDistances(V,beg,S);
    int u = end;
    VI path(1,u);
    while( u != beg && u != inf ) {
        u = par[u];
        path.push_back(u);
    }
    if(u == inf) path.clear();
    reverse(ALL(path));
    return path;
}

VI Solver::findShortestPath(VVI & V, int beg, int end, VB & was) {
    auto [dst,par] = findDistances(V,beg,was);
    int u = end;
    VI path(1,u);
    while( u != beg && u != inf ) {
        u = par[u];
        path.push_back(u);
    }
    if(u == inf) path.clear();
    reverse(ALL(path));
    return path;
}

// finds distances from node beg to all nodes in graph G[ V \ S ].
// this is used to quickly trim states
pair<VI,VI> Solver::findDistances(VVI & V, int beg) {
    int N = V.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        for( int d : V[v] ) {
            if( dst[d] == inf ) {
                dst[d] = dst[v] + 1;
                par[d] = v;
                neigh.push_back(d);
            }
        }
    }

    return make_pair(dst,par);
}
pair<VI,VI> Solver::findDistances(VVI & V, int beg, Mask S) {
    if( S.popcnt() == 0 ) return findDistances(V,beg);

    int N = V.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        for( int d : V[v] ) {
            if( dst[d] == inf && !S[d] ) {
                dst[d] = dst[v] + 1;
                par[d] = v;
                neigh.push_back(d);
            }
        }
    }

    return make_pair(dst,par);
}

pair<VI,VI> Solver::findDistances(VVI & V, int beg, VB & was) {

    int N = V.size();
    VI dst(N, inf);
    VI par(N,inf);
    dst[beg] = 0; par[beg] = -1;
    VI neigh; neigh.reserve(N);
    neigh.push_back(beg);
    for( int i=0; i<neigh.size(); i++ ) {
        int v = neigh[i];
        for( int d : V[v] ) {
            if( dst[d] == inf && !was[d] ) {
                dst[d] = dst[v] + 1;
                par[d] = v;
                neigh.push_back(d);
            }
        }
    }

    return make_pair(dst,par);
}


void Solver::kernelize() {
    Kernelizer kernelizer(V, revV, L, vs, vt);
    kernelizer.cnf = cnf;
    auto core = kernelizer.createCoreGraph();
    V = core.V;
    revV = core.revV;
    vs = core.vs;
    vt = core.vt;
    N = V.size();
    inputL = L = min(core.L, N-1);

    assert( GraphUtils::isSimple(V) );
    assert( GraphUtils::isSimple(revV) );
}


VBIG Solver::solveBySolver2() {
    Solver2 solver(V, L, vs, vt);
    solver.cnf = cnf;
    return solver.solve();
}

void Solver::readGraph(const bool directed){
    string s;
    int M,a,b;

    bool comment = false;
    auto read_comment = [&]() {
        do {
            cin >> s;
            comment = ( s == "c" );
            if(s == "c") getline(cin,s);
        }while( comment );
    };

    read_comment();
    cin >> s >> N >> M;
    assert( s == "edge" );
    V = VVI(N);
    revV = V;

    for(int i = 0; i<M; i++) {
        read_comment();
        assert(s == "e" || s == "a");
        cin >> a >> b;
        if( a == b ) continue; // do not allow loops
        V[a-1].push_back(b-1);
        // if(!directed) V[b-1].push_back(a-1);
        if( s == "e" ) V[b-1].push_back(a-1);
    }

    read_comment();
    assert(s == "l");
    cin >> L;
    L = min(L,N-1);
    inputL = L;

    read_comment();
    assert(s == "t");
    cin >> vs >> vt;

    vs--; vt--;

    for( int i=0; i<N; i++ ) for(int d : V[i]) revV[d].push_back(i);

    clog << "Read graph with " << N << " nodes, " << M << " edges, vs: "
        << vs << ", vt: " << vt << ", L: " << L << endl;
}



// finds D1 and D2 parameters. Also modifies the L value, so that at the end it will hold:
// D1 + D2 + L == input_L_value
void Solver::findD1D2L(VVI & V, VVI & revV, int vs, int vt, int & L, int & D1, int & D2) {
    auto [dst_st_tab,par] = findDistances(V, vs);
    dst_st = dst_st_tab[vt];
    clog << "Nodes " << vs << " and " << vt << " are at distance " << dst_st << endl;
    int K = (L+3) / 4;
    // int K = (L+4) / 3;
    D1 = min( K, dst_st >> 1  );
    D2 = min( K, (dst_st-1) >> 1 );
    L -= D1 + D2;
    if(dst_st == inf) return;

    clog << "D1: " << D1 << ", D2: " << D2 << ", L: " << L << endl;
    clog << "Shortest path from " << vs << " to " << vt << ":" << endl;
    int u = vt;
    VI path(1,u);
    do {
        u = par[u];
        path.push_back(u);
    }while(u != vs);
    reverse(ALL(path));
    for(int d : path) clog << d << " -> ";
    clog << "XXX" << endl;
}

VBIG Solver::countShortestPaths(VVI & V, int vs, VB & was) {
    assert(!was[vs]);
    int N = V.size();

    VI dst(N,inf);
    dst[vs] = 0;
    auto cmp = [&](int a, int b){ if(dst[a] != dst[b]) return dst[a] < dst[b]; return a < b; };
    set<int, decltype(cmp)> zb(cmp);
    for(int i=0; i<N; i++) if(!was[i]) zb.insert(i);

    VBIG res(N,0);
    res[vs] = 1;

    while( !zb.empty() ) {
        int v = *zb.begin();
        zb.erase(zb.begin());

        for( int d : V[v] ) {
            if(was[d]) continue;

            if( dst[v] + 1 < dst[d] ) {
                zb.erase(d);
                dst[d] = dst[v] + 1;
                zb.insert(d);

                res[d] = 0;
            }

            if( dst[v] + 1 == dst[d] ) res[d] += res[v] * 1;
        }
    }

    return res;
}

// checks if exists path from a to b when not using nodes from S.
// this is just for assertions, it is very slow approach
bool Solver::existsPath( VVI & V, int a, int b, Mask & S ) {
    VB was(V.size());
    function<bool(int)> dfs = [&](int num) {
        was[num] = true;
        if( num == b ) return true;
        for(int d : V[num]) if( !was[d] && !S[d] && dfs(d) ) return true;
        return false;
    };
    return dfs(a);
}

// trims states. A state (S, beg, end) is redundant if dst(vs,beg) + |S|-1 + dst(end,vt) > inputL in graph G[V \ S].
void Solver::trimStates(VVI & V, VVI & revV, int vs, int vt, vector<SubsetPaths> & sps) {
    clog << "Trimming states" << endl;
    if( sps.empty() ) {
        clog << "\tnothing to trim! returning..." << endl;
        return;
    }

    clog << "Sorting sps" << endl;
    clog << "All sps: " << endl;
    for(auto sp : sps) clog << sp.toString() << ", mask.size(): " << sp.mask.size()
            << ", sp.mask.m.size(): " << sp.mask.m.size() << endl;
    sort(ALL(sps), [&]( auto & sp1, auto & sp2 ) { return sp1.mask < sp2.mask; } );
    clog << "\tSorted!" << endl;

    VB indices_to_trim( sps.size() );
    int p = 0, q = 0;
    while( p < sps.size() ) {
        auto mask = sps[p].mask;
        clog << "p: " << p << ", -> " << 100.0 * p / sps.size() << endl;
        DEBUG(sps[p].mask.size());
        q = p+1;
        while( q < sps.size() && sps[q].mask == mask) q++;

        // now all SPs from sos[p] to sps[q-1] will have the same mask
        auto [dsts, pars] = findDistances( V, vs, mask );

        // swap(V, revV);
        auto [dstt, part] = findDistances( revV, vt, mask );
        // swap(V, revV);

        // clog << "entries from " << p << " to " << q-1 << " have the same mask:" << endl;
        // clog << mask.toString(V.size()) << endl;
        // for( int i=p; i<q; i++ ) clog << "\t" << sps[i].toString() << endl;

        while(p < q) {

            auto & sp = sps[p];
            int a = sp.beg;
            int b = sp.end;
            int l = sp.length();

            int dstsa = dsts[a];
            for( int d : revV[a] ) dstsa = min( dstsa, 1 + dsts[d] );

            int dsttb = dstt[b];
            for( int d : V[b] ) dsttb = min( dsttb, 1 + dstt[d] );

            if( dstsa + l + dsttb > inputL ) {
                // clog << "State sp: " << sp.toString() << " can be trimmed, dsts: "
                    // << dstsa << ", dsttb: " << dsttb << endl;
                indices_to_trim[p] = true;
            }

            p++;
        }
    }

    for( int i=0; i<sps.size(); i++ ) {
        auto & sp = sps[i];
        if( sp.mask[vs] && vs != sp.beg ) {
            indices_to_trim[i] = true;
            // clog << "Marking sp: " << sp.toString() << " to remove" << endl;
        }
        if( sp.mask[vt] && vt != sp.end ) {
            indices_to_trim[i] = true;
            // clog << "Marking sp: " << sp.toString() << " to remove" << endl;
        }
    }

    for( int i=(int)sps.size()-1; i>=0; i-- ) {
        if( indices_to_trim[i] ) {
            swap( sps[i], sps.back() );
            sps.pop_back();
        }
    }
}


vector<SP> Solver::createNewEntries(int length, int k1, int k2) {
    vector<SubsetPaths> new_entries;

    for( int a=0; a<N; a++ ) sort(ALL(dp[a]), SubsetPaths::sortLBM);

    for( int a = 0; a < N; a++ ) {

        auto [it1a,it2a] = getDPRange( dp[a], k1, [](auto & sp){ return sp.length(); } );

        for( ; it1a != it2a; it1a++ ) {
            assert( it1a->beg == a );
            assert( it1a->length() == k1 );
            auto x = it1a->end;
            assert(x != a);
            auto [it1x,it2x] = getDPRange( dp[x], k2, [](auto & sp){ return sp.length(); } );

            for( ; it1x != it2x; it1x++ ) {
                assert( it1x->beg == x );
                assert( it1x->length() == k2 );

                // if SPs a->x and x->b have a common node other than x, we can continue...
                // if( (it1a->mask & it1x->mask).popcnt() > 1 ) continue;
                {
                    auto newmask = it1a->mask;
                    newmask &= it1x->mask;
                    if( newmask.popcnt() > 1 ) continue;
                }

                assert( it1a->mask[x] && it1x->mask[x] );

                // auto newmask = (it1a->mask | it1x->mask);
                auto newmask = it1a->mask;
                newmask |= it1x->mask;
                HashType newhash = it1x->hash ^ it1x->hash ^ node_hashes[x];
                int b = it1x->end;
                auto newcnt = it1a->counts * it1x->counts;

                new_entries.emplace_back( newmask, length, newhash, a,b, newcnt );

                // clog << "Creating new SPs from it1a: " << it1a->toString() << " and it1x: " << it1x->toString()
                //     << " --- > " << new_entries.back().toString() << endl;
            }
        }
    }

    return new_entries;
}

vector<SP> Solver::createNewEntries2(VVI & V, auto & dp, int length, VI & nodes) {
    vector<SubsetPaths> new_entries;
    new_entries.reserve(1e3);

    // for( int a = 0; a < N; a++ ) {
    for( int a : nodes ) {

        sort(ALL(dp[a]), SubsetPaths::sortLBM);
        auto [it1a,it2a] = getDPRange( dp[a], length-1, [](auto & sp){ return sp.length(); } );

        // clog << "a: " << a << endl;
        // clog << "dp[a]: "; for(auto sp : dp[a]) clog << sp.toString() << endl;

        for( ; it1a != it2a; ++it1a ) {
            // clog << "\tit1a: " << it1a->toString() << endl;
            assert( it1a->beg == a );
            assert( it1a->length() == length-1 );
            auto x = it1a->end;
            assert(x != a);

            for( int d : V[x] ) {
                if( it1a->mask[d] ) continue;

                auto newsp = (*it1a);
                newsp.mask.set(d,true);
                newsp.hash ^= node_hashes[d];
                newsp.l++;
                newsp.end = d;
                new_entries.push_back(newsp);
            }
        }
    }

    return new_entries;

}

// created dp array. After that all SubsetPaths entries with path length not exceeding (L+1) / 2
// will be created
void Solver::createPaths(int max_length, VVI & V, VVI & revV, int vs, int vt, VI nodes, auto & dp) {
    // int K = (L+1) / 2;
    // K = max( K, max(D1,D2) );
    // max_created_path_length = K;
    clog << "Creating paths of lengths up to " << max_length << " for " << nodes.size() << " nodes" << endl;
    // clog << "Creating paths of lengths up to " << max_length << " for " << nodes.size() << " nodes: " << nodes << endl;

    // creating all paths of length 1
    // for( int i=0; i<N; i++ ) {
    for( int i : nodes ) {
        if( i == vt ) continue;
        for( int d : V[i] ) {
            if( d == vs ) continue;

            Mask m = Mask::zeros(N);
            m.set(i,true);
            m.set(d,true);
            auto hash = HashType( node_hashes[i].hash ^ node_hashes[d].hash );

            SP sp(m, 1, hash, i,d, 1);
            dp[i].push_back(sp);
        }
    }


    // for( int length = 2; length <= K; length++ ) {
    for( int length = 2; length <= max_length; length++ ) {

        // int k1 = length / 2;
        // int k2 = length - k1;

        // clog << endl << "********************* Creating paths of length: " << length << ", k1: " << k1
            // << ", k2: " << k2 << ", nodes.size(): " << nodes.size() << endl;


        clog << endl << "********************* Creating paths of length: " << length
             << ", nodes.size(): " << nodes.size() << endl;


        // auto new_entries = createNewEntries(length, k1, k2);
        auto new_entries = createNewEntries2(V,dp,length,nodes);
        // clog << "\tnew entries: " << endl; for(auto sp : new_entries) clog << "\t" << sp.toString() << endl;

        // now all new entries should be created, it is time to merge then to save some space...
        // clog << "Sorting entries" << endl;
        sort(ALL(new_entries), SP::sortABLM);

        int p = 0, q = 0, r = 0;
        constexpr bool use_merging = true;
        if constexpr (use_merging) {
            while( p < new_entries.size() ){
                q = p;
                auto sp1 = new_entries[p], sp2 = new_entries[q];
                while( q < new_entries.size() && sp1 == sp2 ) {
                    q++;
                    if(q <new_entries.size()) sp2 = new_entries[q];
                }

                while( p+1 < q ) { // we can merge entries, all will be merged to sp1
                    sp1.counts += new_entries[++p].counts;
                    // clog << "Afteer merging, sp1: " << sp1.toString() << endl;
                }

                new_entries[r++] = sp1;
                p = q;
            }
        }else r = new_entries.size();

        clog << "Before merging states, new_entries.size(): " << new_entries.size() << endl;
        clog << "entries merged: " << new_entries.size() - r << endl;
        // new_entries.resize(r);
        while(new_entries.size() > r) new_entries.pop_back();
        clog << "After merging states, new_entries.size(): " << new_entries.size() << endl;

        if(new_entries.size() < 1e7) // #TEST - trim entries only if there are not so many of them...
            trimStates(V, revV, vs, vt, new_entries);

        for( auto & sp : new_entries ) dp[sp.beg].push_back( sp );
        clog << "After trimming, there are " << new_entries.size() << " new entries" << endl;

    }
}


VBIG Solver::calculateResultSpecialCase1(VVI & V, VVI & revV, auto & dp, auto & dp_rev, int vs, int vt) {
    assert( D1 == 1 && D2 == 0 );
    // bigint res = 0;

    for( int a=0; a<N; a++ )  sort( ALL(dp[a]), SP::sortBLM );

    // we consider all paths a -> x -> vt, where a is a neighbor of vs.

    vector<bigint> res_for_length(N,0);

    for( int a : V[vs] ) {
        // clog << "a: " << a << endl;
        for( auto & sp : dp[a] ) {

            int x = sp.end;
            if( x == vt && sp.length() == 1  ) {
                // clog << "Found path " << a << "  -->  " << sp.toString() << ", path length: " << 1 + sp.length() << endl;
                // if( sp.length() > 1 ) res_for_length[1+sp.length()] += sp.counts;
                // else
                //     res += sp.counts;
                res_for_length[1 + sp.length()] += sp.counts;
                continue;
            }

            auto [it1, it2] = getDPRange( dp[x], vt, []( SP & spt ){ return spt.end; } );
            for( ; it1 != it2; it1++ ) {
                // Mask mask_inters = (sp.mask & it1->mask);
                Mask mask_inters = sp.mask;
                mask_inters &= it1->mask;
                assert( mask_inters[x] );
                assert( it1->end == vt );

                if( mask_inters.popcnt() == 1 ) {
                    auto length = 1 + sp.length() + it1->length();
                    if(length > inputL) continue;

                    res_for_length[length] += sp.counts * it1->counts;
                    // clog << "\tFound paths " << a << "  -->  " << sp.toString() << "  -->  " << it1->toString()
                        // << ", path length: " << length << endl;
                }
            }
        }
    }

    for( int l=3; l<=inputL; l++ ) {
        int duplicates_x = min( l-2, 2*max_created_path_length - l + 2 );
        if( res_for_length[l] == 0 ) continue;
        // clog << "res_for_length[" << l << "]: " << res_for_length[l] << endl;
        // clog << "duplicates_x: " << duplicates_x << endl;
        assert( res_for_length[l] % duplicates_x == 0 );
        // res += res_for_length[l] / duplicates_x;
    }

    return res_for_length;
}


VBIG Solver::calculateResult(VVI & V, VVI & revV, auto & dp, auto & dp_rev, int vs, int vt) {
    constexpr bool debug = false;
    clog << "Starting calculation of result" << endl;

    if( D1 == 1 && D2 == 0 ) {
        assert(false && "calculateREsultSpecialCase() should never be called since, the MIM method should do its work");
        clog << "Solving special case!" << endl;
        return calculateResultSpecialCase1(V, revV, dp, dp_rev, vs, vt);
    }

    vector<bigint> res_for_length(N,0);

    if(debug) {
        clog << "Before removing entries from dp[vs] and dp_rev[vt]:" << endl;
        clog << "dp[vs].size(): " << dp[vs].size() << endl;
        for(auto & sp : dp[vs]) clog << sp.toString() << endl;
        clog << endl;
        clog << "dp_rev[vt].size(): " << dp_rev[vt].size() << endl;
        for(auto & sp : dp_rev[vt]) clog << sp.toString() << endl;
    }


    { // remove from dp[vs] all entries that have length != D1 and from dp_rev[vt] that have length != D2
        for( int i=(int)dp[vs].size()-1; i>=0; i-- ) {
            auto & sp = dp[vs][i];
            if( sp.length() != D1 ) {
                swap( dp[vs][i], dp[vs].back() );
                dp[vs].pop_back();
            }
        }

        for( int i=(int)dp_rev[vt].size()-1; i>=0; i-- ) {
            auto & sp = dp_rev[vt][i];
            if( sp.length() != D2 ) {
                swap( dp_rev[vt][i], dp_rev[vt].back() );
                dp_rev[vt].pop_back();
            }
        }
    }

    if(debug) {
        clog << "After removing entries from dp[vs] and dp_rev[vt]:" << endl;
        clog << "dp[vs].size(): " << dp[vs].size() << endl;
        for(auto & sp : dp[vs]) clog << sp.toString() << endl;
        clog << "dp_rev[vt].size(): " << dp_rev[vt].size() << endl;
        for(auto & sp : dp_rev[vt]) clog << sp.toString() << endl;
    }


    for( int a=0; a<N; a++ ) {
        sort( ALL(dp[a]), SP::sortBLM );
        sort( ALL(dp_rev[a]), SP::sortBLM );
    }

    set<int> zbA;
    for( auto & sp : dp[vs] ) zbA.insert(sp.end);
    VI A(ALL(zbA));

    set<int> zbB;
    for( auto & sp : dp_rev[vt] ) zbB.insert(sp.end);
    VI B(ALL(zbB));


    if(debug) {
        clog << "A: "; for(int a : A) clog << a << " "; clog << endl;
        clog << "B: "; for(int b : B) clog << b << " "; clog << endl;
    }

    // now we can proceed to calculations...
    // for each a in A and b in B we count paths vs -> a -> x -> b -> vt

    clog << endl << "A.size(): " << A.size() << endl << "B.size(): " << B.size() << endl;

    int progress_cnt = 0;
    for( int a : A ) {
        clog << "\rprogress: " << ( (double)progress_cnt++ / A.size() * 100 ) << "%" << flush;

        if(debug) {
            clog << "a: " << a << endl;
            clog << "dp[a]: "; for( auto & sp : dp[a] ) clog << sp.toString() << endl;
        }

        // admissible_sps_vs[j] is the set of SPs such that thir intersection with dp_rev[b][j]
        // is empty and the end node is a. Similarly for admissible_sps_vt.
        // creating admissible_sps_vt can be done here and only slightly modified for give b, instead of creating anew
        vector<vector<SP>> admissible_sps_vt(dp[a].size());
        for( int i=0; i < dp[a].size(); i++ ) {
            for( auto & sp : dp_rev[vt] ) {
                // auto mask = ( dp[a][i].mask & sp.mask );
                auto mask = dp[a][i].mask;
                mask &= sp.mask;
                if( mask.popcnt() == 0 ) admissible_sps_vt[i].push_back(sp);
            }
        }

        for( int b : B ) {
            if(debug) {
                clog << "\tb: " << b << endl;
                clog << "\tdp_rev[b]: "; for( auto & sp : dp_rev[b] ) clog << "\t" << sp.toString() << endl;
            }

            vector<pair<int,SP>> sps_removed; // we need to remove those entries that have end in nodes other than b
            for( int i=0; i<dp[a].size(); i++ ) {
                for(int j=(int)admissible_sps_vt[i].size()-1; j>=0; j--) {
                    auto & sp = admissible_sps_vt[i][j];
                    if( sp.end != b ) {
                        sps_removed.emplace_back(i,sp);
                        swap( admissible_sps_vt[i][j], admissible_sps_vt[i].back() );
                        admissible_sps_vt[i].pop_back();
                    }
                }

                // clog << "admissible_sps_vt[" << i << "].size(): " << admissible_sps_vt[i].size() << endl;
            }

            { // now the first case, where b is a direct neighbor of a
                if( find(ALL(V[a]),b) != V[a].end() ) {
                    if(debug) clog << "\ttThere are connections vs->a->b->vt without intermediate node x" << endl;

                    bigint cnt_vs = 0;
                    for( auto & sp : dp[vs] ) {
                        if( sp.end == a ) {
                            cnt_vs += sp.counts;
                            assert( sp.length() == D1 );
                        }

                    }

                    bigint cnt_vt = 0;
                    for( auto & sp : dp_rev[vt] ) {
                        if( sp.end == b ) {
                            cnt_vt += sp.counts;
                            assert( sp.length() == D2 );
                        }
                    }
                    if(debug) clog << "\t\tcnt_vs: " << cnt_vs << ", cnt_vt: " << cnt_vt << endl;

                    res_for_length[D1+D2+1] += cnt_vs * cnt_vt;
                }
            }


            // #TEST - these are just counters for performance analysis, they can be removed for production
            int adm_disp_copies = 0;
            int total_checks = 0;
            int valid_checks = 0;
            int total_adm_checks = 0;
            int valid_adm_checks = 0;
            int cached_values_hits = 0;

            // admissible_sps_vs[j] is the set of SPs such that thir intersection with dp_rev[b][j]
            // is empty and the end node is a. Similarly for admissible_sps_vt.
            vector<vector<SP>> admissible_sps_vs(dp_rev[b].size());


            for( int j=0; j<dp_rev[b].size(); j++ ) {
                for( auto & sp : dp[vs] ) {
                    adm_disp_copies++;
                    if( sp.end != a ) continue;
                    // auto mask = ( dp_rev[b][j].mask & sp.mask );
                    auto mask = dp_rev[b][j].mask;
                    mask &= sp.mask;
                    if( mask.popcnt() == 0 ) admissible_sps_vs[j].push_back(sp);
                }
            }

            int pa = 0, pb = 0, qa = 0, qb = 0;
            map<HashType,bigint> cached_values;

            while( pa < dp[a].size() ) {
                qa = pa;
                while( qa < dp[a].size() && dp[a][qa].end == dp[a][pa].end ) qa++;
                // now all entries in range dp[a][pa : qa] (without index qa) have the same end x
                int x = dp[a][pa].end;

                while( pb < dp_rev[b].size() && dp_rev[b][pb].end < x ) pb++;
                qb = pb;
                while( qb < dp_rev[b].size() && dp_rev[b][qb].end == x ) qb++;
                // now all entries in range dp_rev[b][pb : qb] (without index qb) have the same end x.
                // it might be that qb = pb and there is no valid entry in dp_rev[b]

                assert( pa < qa ); // this should always hold...
                assert( x != a );
                if( x == b || pb == qb ) { // if there is no valid path from x to b, then we can continue...
                    pa = qa;
                    continue;
                }

                // now we have some set of paths from a to x and some set of paths from x to b.
                // we now need to check (preferably in a smart way) all pairs with masks that intersect only in x

                // clog << "(qa-pa)*(qb-pb): " << (qa-pa) * (qb-pb) << endl;

                for( int i=pa; i < qa; i++ ) {
                    auto & spa = dp[a][i];
                    auto & mask_a = spa.mask;
                    // clog << "here, a: " << a << ", spa: " << spa.toString() << endl;

                    vector<SP> adm_sps_vt; // will use only these admissible sps that have empty intersection with mask_a
                    for( auto & sp : admissible_sps_vt[i] ) {
                        // auto mask_inters = ( sp.mask & mask_a );
                        auto mask_inters = sp.mask;
                        mask_inters &= mask_a;
                        if( mask_inters.popcnt() == 0 ) adm_sps_vt.push_back(sp);
                    }

                    for( int j=pb; j < qb; j++ ) {
                        auto & spb = dp_rev[b][j];
                        int length = spa.length() + spb.length() + D1 + D2;

                        total_checks++;
                        if( length > inputL ) continue;
                        // clog << "\tb: " << b << ", spb: " << spb.toString() << endl;

                        auto & mask_b = spb.mask;

                        // auto mask_inters = (mask_a & mask_b);
                        auto mask_inters = mask_a;
                        mask_inters &= mask_b;
                        assert( mask_inters[x] ); // x must belong to both spa and spb

                        if( mask_inters.popcnt() > 1 ) continue;
                        valid_checks++;

                        // auto mask_ab = ( mask_a | mask_b );
                        auto mask_ab = mask_a;
                        mask_ab |= mask_b;

                        auto state_hash = spa.hash ^ spb.hash ^ node_hashes[x];

                        // now we can find all masks from SPs in dp[vs] that have only common node a with mask_ab
                        // it suffices to iterate over only SPs in admissible_sps_vs[j]
                        // we calculate the sum of counts of those sps...
                        bigint cnt_vs = 0;
                        if( cached_values.contains( state_hash ) ) {
                            res_for_length[ length ] += spa.counts * spb.counts * cached_values[state_hash];
                            cached_values_hits++;
                        }
                        else {
                            for( auto & sp : admissible_sps_vs[j] ) {
                                // mask_inters = ( sp.mask & mask_ab );
                                mask_inters = sp.mask;
                                mask_inters &= mask_ab;

                                // if( !mask_inters[a] ) continue;
                                assert(mask_inters[a]); // node a must be in intersection
                                total_adm_checks++;
                                if( mask_inters.popcnt() == 1 ) {
                                    valid_adm_checks++;
                                    cnt_vs += sp.counts; // we found a valid path vs->a->b

                                    if(debug) {
                                        swap( spb.beg, spb.end );
                                        clog << "\t\tFound valid path vs->a->x->b: " << endl;
                                        clog << "\t\t" << sp.toString() << "  -->  " << spa.toString() << "  -->  " << x << "  -->  "
                                            << spb.toString() << endl;
                                        swap( spb.beg, spb.end );
                                    }
                                }
                            }

                            bigint cnt_vt = 0;
                            // for( auto & sp : admissible_sps_vt[i] ) {
                            for( auto & sp : adm_sps_vt ) {
                                // mask_inters = ( sp.mask & mask_ab );
                                mask_inters = sp.mask;
                                mask_inters &= mask_ab;

                                // if( !mask_inters[b] ) continue;
                                assert(mask_inters[b]); // node b must be in intersection
                                total_adm_checks++;
                                if( mask_inters.popcnt() == 1 ) {
                                    valid_adm_checks++;
                                    cnt_vt += sp.counts; // we found a valid path a->b->vt

                                    if(debug) {
                                        swap(sp.beg, sp.end); swap( spb.beg, spb.end );
                                        clog << "\t\tFound valid path a->x->b->vt: " << endl;
                                        clog << "\t\t" << spa.toString() << "  -->  " << x << "  -->  " << spb.toString() << "  -->  "
                                            << sp.toString() << endl;
                                        swap(sp.beg, sp.end); swap( spb.beg, spb.end );
                                    }
                                }
                            }

                            if(debug) {
                                clog << "\t\tpath length: " << length << endl;
                                clog << "\t\tFound cnt_vs: " << cnt_vs << " valid paths vs->a->b" << endl;
                                clog << "\t\tFound cnt_vt: " << cnt_vt << " valid paths a->b->vt" << endl;
                                clog << "\t\tspa.counts: " << spa.counts << ", spb.counts: " << spb.counts << endl;
                            }

                            cached_values[state_hash] = cnt_vs * cnt_vt;
                            res_for_length[ length ] += spa.counts * spb.counts * cnt_vs * cnt_vt;
                        }

                    }
                }

                pa = qa;
            }

            // clog << "total_checks: " << total_checks << " from which only " << valid_checks << " had nonintersecting mask" << endl;
            // clog << "total admissible checks: " << total_adm_checks << ", from which nonitersecting: " << valid_adm_checks << endl << endl;
            // clog << "adm_disp_copies: " << adm_disp_copies << endl;
            // clog << "cached_values_hits: " << cached_values_hits << endl;

            // now we need to put removed nodes back to admissible_sps_vt
            for( auto [i,sp] : sps_removed ) admissible_sps_vt[i].push_back(sp);
        }
    }

    if(debug) clog << "max_created_path__length: " << max_created_path_length << endl;


    // bigint res = res_for_length[D1+D2+1];

    for( int l=D1+D2+2; l<=inputL; l++ ) {
        int duplicates_x = min( (l+1 - (D1+1) - (D2+1)), 2*max_created_path_length - (l+1 - (D1+1) - (D2+1)) );

        if(debug) {
            clog << "res_for_length[" << l << "]: " << res_for_length[l] << endl;
            clog << "'loose' nodes on path: " << (l+1 - (D1+1) - (D2+1)) << endl;
            clog << "path of that length shoould be counted " << duplicates_x << " times" << endl;
        }
        if( res_for_length[l] == 0 ) continue;

        assert( res_for_length[l] % duplicates_x == 0 );
        // res += res_for_length[l] / duplicates_x;
    }

    return res_for_length;
}

VBIG Solver::solveSmartMIM() {
    clog << "Solving using meet-in-the-middle" << endl;
    if( N > MAXN ) {
        cerr << "CANNOT PROCESS GRAPHS WITH MORE THAN " << MAXN << " NODES USING MIM..." << endl;
        exit(1);
    }

    vector<bigint> res_for_length(N);

    bool has_vsvt_arc = false;
    if( StandardUtils::find( V[vs],vt ) ) has_vsvt_arc = true;
    if( has_vsvt_arc ) {
        Kernelizer::removeArc(V, revV, vs,vt);
        Kernelizer::removeArc(V, revV, vt,vs);
        res_for_length[1] += 1;
    }

    for( int i=0; i<N; i++ ) {
        dp[i].clear();
        dp_rev[i].clear();
    }

    int max_length = (inputL+1)/2;

    createPaths( max_length, V, revV, vs, vt, {vs}, dp );
    createPaths( max_length, revV, V, vt, vs, {vt}, dp_rev );

    int pa = 0, pb = 0, qa = 0, qb = 0;


    // clog << "dp[vs]: " << endl; for(auto & sp : dp[vs]) clog << sp.toString() << endl;
    // clog << "dp_rev[vt]: " << endl; for(auto & sp : dp_rev[vt]) clog << sp.toString() << endl;

    sort(ALL(dp[vs]), SP::sortBLM);
    sort(ALL(dp_rev[vt]), SP::sortBLM);

    while( pa < dp[vs].size() ) {
        clog << "\rMim progress: " << int(( 1.0*pa / dp[vs].size() ) * 100) << "%" << flush;

        qa = pa;
        while( qa < dp[vs].size() && dp[vs][qa].end == dp[vs][pa].end ) qa++;
        // now all entries in range dp[a][pa : qa] (without index qa) have the same end x
        int x = dp[vs][pa].end;

        while( pb < dp_rev[vt].size() && dp_rev[vt][pb].end < x ) pb++;
        qb = pb;
        while( qb < dp_rev[vt].size() && dp_rev[vt][qb].end == x ) qb++;
        // now all entries in range dp_rev[b][pb : qb] (without index qb) have the same end x.
        // it might be that qb = pb and there is no valid entry in dp_rev[b]

        assert( pa < qa ); // this should always hold...
        if( pb == qb ) { // if there is no valid path from x to b, then we can continue...
            pa = qa;
            continue;
        }

        // now we have some set of paths from a to x and some set of paths from x to b.
        // we now need to check (preferably in a smart way) all pairs with masks that intersect only in x

        for( int i=pa; i < qa; i++ ) {
            auto & spa = dp[vs][i];
            auto & mask_a = spa.mask;
            // clog << "sp_vs: " << spa.toString() << endl;

            // we split by middle point, so we can check only spb's for which length is equal or 1 less than given spa
            while( pb < qb && dp_rev[vt][pb].length()+1 < spa.length() ) pb++;
            if( 2*spa.length() -1 > inputL ) break;

            for( int j=pb; j < qb; j++ ) {
                auto & spb = dp_rev[vt][j];
                if( spb.length() > spa.length() || spa.length() + spb.length() > inputL ) break;

                // auto mask_inters = (mask_a & spb.mask);
                auto mask_inters = mask_a;
                mask_inters &= spb.mask;
                // assert( mask_inters[x] ); // x must belong to both spa and spb

                // if( mask_inters.popcnt() > 1 ) continue;

                // clog << "\tFound path of length " << length << ": " << spa.toString() << " --> " << spb.toString() << endl;
                res_for_length[spa.length() + spb.length()] += (mask_inters.popcnt() <= 1 ? spa.counts * spb.counts : (bigint)0);
            }
        }

        pa = qa;
    }

    return res_for_length;
}


VBIG Solver::solveMIM2() {
    constexpr bool debug = false;

     clog << "Solving using solver1.mim2" << endl;
    if( N > MAXN ) {
        cerr << "CANNOT PROCESS GRAPHS WITH MORE THAN 256 NODES USING MIM..." << endl;
        exit(1);
    }

    int max_length = (L+1)/2;

    vector<bigint> res_for_length(L+2);
    bool has_vsvt_arc = false;
    if( StandardUtils::find( V[vs],vt ) ) has_vsvt_arc = true;
    if( has_vsvt_arc ) {
        Kernelizer::removeArc(V, revV, vs,vt);
        Kernelizer::removeArc(V, revV, vt,vs);
        res_for_length[1] += 1;
    }


    vector<vector<SP3>> dpvs1(N), dpvs2(N), dpvt1(N), dpvt2(N);

    dpvs1[vs].emplace_back( Mask(N), node_hashes[vs], 1 );
    dpvs1[vs].back().mask.set(vs,1);

    dpvt1[vt].emplace_back( Mask(N), node_hashes[vt], 1 );
    dpvt1[vt].back().mask.set(vt,1);

    auto mergeEntries = [&](auto & new_entries) {
        int p = 0, q = 0, r = 0;
        while( p < new_entries.size() ){
            q = p;
            auto sp1 = new_entries[p], sp2 = new_entries[q];
            while( q < new_entries.size() && (sp1 == sp2) ) if(++q <new_entries.size()) sp2 = new_entries[q];
            while( p+1 < q ) sp1.counts += new_entries[++p].counts;
            new_entries[r++] = sp1;
            p = q;
        }

        // new_entries.resize(r);
        while(new_entries.size() > r) new_entries.pop_back();
    };

    auto trimStates = [L=L]( VVI & V, int source, int v, auto & sps ) {
        sort(ALL(sps) );

        VB indices_to_trim( sps.size() );
        int p = 0;
        while( p < sps.size() ) {
            auto& mask = sps[p].mask;
            assert(mask[v]);
            mask.set(v,0);
            auto [dst, par] = findDistances( V, source, mask );
            mask.set(v,1);
            if( mask.popcnt()-1 + dst[v] > L ) indices_to_trim[p] = true;
            p++;
        }

        for( int i=(int)sps.size()-1; i>=0; i-- ) {
            if( indices_to_trim[i] ) {
                if constexpr (debug) clog << "Trimming state: " << sps[i].toString() << endl;
                swap( sps[i], sps.back() );
                sps.pop_back();
            }
        }
    };


    for( int l=1; l <= max_length; l++ ) {

        // if constexpr (debug)
            clog << endl << endl << "Considering length: " << l << endl;

        // now increase path length by 1
        swap(dpvs1, dpvs2);
        swap( dpvt1, dpvt2 );

        for( int i=0; i<N; i++ ) {
            dpvs1[i].clear();
            dpvt1[i].clear();
        }

        for( int i=0; i<N; i++ ) { // create in dpvs1 and dpvt1 entries with new lengths
            for( auto & sp3 : dpvs2[i] ) {
                for( int d : V[i] ) {
                    if( !sp3.mask[d] && d != vs && d != vt ) {
                        dpvs1[d].emplace_back( sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts );
                        dpvs1[d].back().mask.set(d,1);
                    }
                }
            }

            for( auto & sp3 : dpvt2[i] ) {
                for(int d : revV[i]) {
                    if( !sp3.mask[d] && d != vs && d != vt ) {
                        dpvt1[d].emplace_back( sp3.mask, sp3.hash ^ node_hashes[d], sp3.counts );
                        dpvt1[d].back().mask.set(d,1);
                    }
                }
            }
        }

        int total_sps = 0;
        for(int i=0; i<N; i++) total_sps += dpvs1[i].size() + dpvt1[i].size() + dpvt2[i].size();
        clog << "\tTotal sps: " << total_sps << endl;

        if constexpr (debug) {
            clog << "Created SP3s:" << endl;

            for(int i=0; i<N; i++) {
                if( dpvs1[i].empty() ) continue;
                clog << "dpvs1[" << i << "]: " << endl;
                for( auto & sp : dpvs1[i] ) clog << "\t" << sp.toString() << endl;
            }

            for(int i=0; i<N; i++) {
                if( dpvt1[i].empty() ) continue;
                clog << "dpvt1[" << i << "]: " << endl;
                for( auto & sp : dpvt1[i] ) clog << "\t" << sp.toString() << endl;
            }

        }

        if constexpr (debug) clog << "\tProceeding to merge entries and trim states" << endl;
        for( int i=0; i<N; i++ ) {
            sort(ALL(dpvs1[i]));
            mergeEntries( dpvs1[i] );
            trimStates( revV, vt, i, dpvs1[i] );

            sort(ALL(dpvt1[i]));
            mergeEntries( dpvt1[i] );
            trimStates( V, vs,i, dpvt1[i] );
        }

        total_sps = 0;
        for(int i=0; i<N; i++) total_sps += dpvs1[i].size() + dpvt1[i].size() + dpvt2[i].size();
        clog << "\tAfter merging and trimming, total sps: " << total_sps << endl;

        // finally, calculate results...
        if constexpr (debug) clog << "\tProceeding to calculate results for lengths " << 2*l-1 << " and " << 2*l << endl;
        ULL total_comparisons = 0;
        for( int i=0; i<N; i++ ) total_comparisons += dpvs1[i].size() * ( dpvt1[i].size() + dpvt2[i].size() );
        clog << "\tThere needs to be done " << total_comparisons << " bitset mask comparisons" << endl;

        for( int i=0; i<N; i++ ) {

            for( auto & sp1 : dpvs1[i] ) {
                bigint res = 0;

                if( 2*l <= L ) {
                    for( auto & sp2 : dpvt1[i] ) {
                        // if( (sp1.mask & sp2.mask).popcnt() == 1 ) {
                        //     if constexpr (debug) clog << "\tFound path: " << sp1.toString() << " -> " << sp2.toString() << endl;
                        //     res += sp2.counts;
                        // }
                        res += ( sp1.mask.singleNodeIntersection(sp2.mask)? sp2.counts : 0);
                    }
                    res_for_length[2*l] += sp1.counts * res;
                }

                if( 2*l-1 <= L ) {
                    res = 0;
                    for( auto & sp2 : dpvt2[i] ) {
                        // if( (sp1.mask & sp2.mask).popcnt() == 1 ) {
                        //     if constexpr (debug) clog << "\tFound path: " << sp1.toString() << " -> " << sp2.toString() << endl;
                        //     res += sp2.counts;
                        // }
                        // res += ( (sp1.mask & sp2.mask).popcnt() == 1 ? sp2.counts : 0);
                        res += ( sp1.mask.singleNodeIntersection(sp2.mask) ? sp2.counts : 0);
                    }
                    res_for_length[2*l-1] += sp1.counts * res;
                }
            }
        }
    }

    while( res_for_length.size() > L+1 ) res_for_length.pop_back();
    return res_for_length;
}


VBIG Solver::solveSmart() {
    Mask mask(N);
    if( !existsPath(V,vs,vt,mask) ) return VBIG(L+1,0);

    bool has_st_arc = (find( ALL(V[vs]), vt ) != V[vs].end());

    VBIG res_for_length(L+1,0);
    res_for_length[1] = (has_st_arc ? 1 : 0);

    {
        // remove vs->vt arc and vt->vs ar, if present
        auto removeNodeFromList = [](VI & v, int nd){
            for( int i=0; i<v.size(); i++ ) {
                if( v[i] == nd ) {
                    swap( v[i], v.back() );
                    v.pop_back();
                }
            }
        };
        removeNodeFromList(V[vs],vt);  removeNodeFromList(revV[vt],vs);
        removeNodeFromList(V[vt],vs);  removeNodeFromList(revV[vs],vt);
    }

    kernelize();

    clog << "After kernelization graph has " << N << " nodes, " << GraphUtils::countEdges(V,true)
        << " arcs and L: " << L << endl;

    if( GraphUtils::countEdges(V,true) == 0 ) return res_for_length;


    // use_graphillion_for_all_paths = false; // #TEST - this might be problematic, as it returns only a single bigint...
    // if( use_graphillion_for_all_paths && L >= N-1 && !GraphUtils::isDirected(V) ) {
    //     // if( use_graphillion_for_all_paths && L >= N-1) { // #TEST
    //     // V = GraphUtils::getUndirectedStructure(V, {}); // #TEST - this is only for tests...
    //     // revV = GraphUtils::reverseGraph(V); // #TEST - this is not correct, only for tests...
    //
    //
    //     string s = StandardUtils::getRandomString(10);
    //     string input_filename = s + ".col";
    //     ofstream f(input_filename);
    //     GraphUtils::writeInstanceToFile( f, V, vs, vt, L, false );
    //     f.close();
    //
    //     string cmd = "python3 countAllPaths.py " + input_filename;
    //     clog << endl << "Counting all paths using graphillion, cmd: " << cmd << endl;
    //     system( cmd.c_str() );
    //
    //     string res_filename = s + ".out";
    //     ifstream str(res_filename);
    //     str >> s;
    //     str.close();
    //
    //     auto cast_to_bigint = [](auto s) { return stoll(s); };
    //
    //     filesystem::remove(input_filename);
    //     filesystem::remove(res_filename);
    //
    //     res_for_length = VBIG(1, cast_to_bigint(s) );
    //
    //     return res_for_length;
    // }

    constexpr bool split_by_dominators = true;
    if constexpr (split_by_dominators){ // splitting into disjoint cases...
        constexpr bool debug = true;

        auto dominators = Kernelizer::findDominators(V,vs,vt);
        if( !dominators.empty() ) {
            if constexpr (debug) DEBUG(N);
            if constexpr (debug) DEBUG(L);
            if constexpr (debug) DEBUG(dominators);
            VB was(N);
            for(int d : dominators) was[d] = true;
            dominators.clear();
            auto path = findShortestPath(V,vs,vt, Mask::zeros(N));
            VI sp_lengths;
            int p = 0;
            for( int i=0; i<path.size(); i++ ) {
                int v = path[i];
                if( was[v] ) { // v is a dominator
                    dominators.push_back(v);
                    sp_lengths.push_back(i-p);
                    p = i;
                }
            }
            for(int d : dominators) was[d] = false;

            sp_lengths.push_back((int)path.size() - 1 - p);
            if constexpr (debug) DEBUG(path.size());
            if constexpr (debug) DEBUG(dominators);
            if constexpr (debug) DEBUG(sp_lengths);

            VI split_nodes = {vs};
            split_nodes += dominators;
            split_nodes.push_back(vt);
            if constexpr (debug) DEBUG(split_nodes);

            if( split_nodes.size() > 2 ) {
                Lengths res( 0,0);
                res[0] = 1;

                int suml = accumulate(ALL(sp_lengths),0);

                if constexpr (debug) clog << "Splitting paths vs -> vt: " << vs << " -> " << vt << " with L: " << L << " into following cases:" << endl;
                for( int i=1; i<split_nodes.size(); i++ ) {
                    int a = split_nodes[i-1], b = split_nodes[i], newL =  L - ( suml - sp_lengths[i-1] );
                    if constexpr (debug) clog << "Path " << a << " -> " << b << " with L: " << newL << endl;

                    was[b] = true;
                    auto [dst,par] = findDistances( V, a, was );
                    VI nodes(1,b);
                    for( int j=0; j<N; j++ ) if( !was[j] && dst[j] != inf ) nodes.push_back(j);
                    for( int d : nodes ) was[d] = true;
                    was[b] = false;

                    if constexpr (debug) clog << "\tInducing graph by nodes: " << nodes << endl;

                    auto indg = GraphInducer::induceDirected( V, revV, nodes );
                    Solver solver( indg.V, newL, indg.perm[a], indg.perm[b] );
                    solver.cnf = cnf;
                    // if constexpr (debug) DEBUG(indg);

                    auto res_indg = solver.solveSmart();
                    if constexpr (debug) DEBUG(res_indg);
                    Lengths lindg( 0, (int)res_indg.size()-1 );
                    for( int j=1; j<res_indg.size(); j++ ) lindg[j] = res_indg[j];
                    res = Lengths::multiplyLengths( res, lindg );
                }

                if constexpr (debug) DEBUG(res);

                for(int i=res.minl(); i <= min(res.maxl(), (int)res_for_length.size()-1); i++) {
                    res_for_length[i] += res[i];
                }
                return res_for_length;
            }

        }else {
            if constexpr (debug) {
                clog << "Graph cannot be split into disjoint cases, there are no dominators..." << endl;
            }
        }
    }



    assert( L == inputL );

    // if( L < 0 ) {
    if( false ) {
        auto res = solveBT();
        for( int i=0; i<min(res_for_length.size(),res.size()); i++ ) res_for_length[i] += res[i];
    }
    // else if( L < 0 ) { // #TEST #CAUTION - just for tests
    else if( false ) { // #TEST #CAUTION - just for tests
        auto res = solveSmartMIM();
        for( int i=0; i<min(res_for_length.size(),res.size()); i++ ) res_for_length[i] += res[i];
    }
    else if( true ) {
        auto res = solveBySolver2();
        for( int i=0; i<min(res_for_length.size(),res.size()); i++ ) res_for_length[i] += res[i];
    }
    else {
        if( N > 256 ) {
            cerr << "CANNOT PROCESS GRAPHS WITH MORE THAN 256 NODES USING MIM..." << endl;
            exit(1);
        }

        // use this to admit using MIM with extension
        // #CAUTION! This changes the value of variable L...
        findD1D2L(V, revV, vs,vt, L, D1, D2);

        clog << "dst_st: " << dst_st << endl;
        if( dst_st == inf ) return res_for_length;
        assert( dst_st >= 2 ); // now dst_st >= 2, because there is no arc vs->vt
        assert( D1 > 0 || D2 > 0 );


        clog << "Creating paths of length up to " << D1 << " for vs" << endl;
        createPaths(D1, V,  revV, vs,vt, {vs}, dp);
        clog << "Creating paths of length up to " << D2 << " for vt" << endl;
        createPaths(D2, revV, V, vt,vs, {vt}, dp_rev);

        for( int a=0; a<N; a++ ) {
            sort( ALL(dp[a]), SP::sortBLM );
            sort( ALL(dp_rev[a]), SP::sortBLM );
        }

        set<int> zbA;
        for( auto & sp : dp[vs] ) zbA.insert(sp.end);
        VI A(ALL(zbA));

        set<int> zbB;
        for( auto & sp : dp_rev[vt] ) zbB.insert(sp.end);
        VI B(ALL(zbB));

        clog << "A: "; for(auto a : A) clog << a << " "; clog << endl;
        clog << "B: "; for(auto b : B) clog << b << " "; clog << endl;

        max_created_path_length = (L+1)/2;
        clog << "Creating paths up to length max_created_path_length: " << max_created_path_length << " for nodes in A, |A| = " << A.size() << endl;
        clog << "Creating paths up to length max_created_path_length: " << max_created_path_length << " for nodes in B, |B| = " << B.size() << endl;
        createPaths( max_created_path_length, V, revV, vs,vt, A, dp );
        createPaths( max_created_path_length, revV, V, vt,vs, B, dp_rev );


        auto res = calculateResult(V, revV, dp, dp_rev, vs, vt);
        for( int i=0; i<res.size(); i++ ) res_for_length[i] += res[i];
    }

    return res_for_length;
}
