//
// Created by sylwe on 12.07.2024.
//

#ifndef SOLVER_H
#define SOLVER_H

#include "Config.h"
#include "Structures.h"

class Solver {
public:
    Solver(VVI _V, int _L, int _vs, int _vt);

    int N; // size of the input graph
    int L; // max length of paths to consider
    int max_created_path_length = -1;
    int inputL;
    int D1,D2; // vicinity distances from nodes vs and vt
    int dst_st = -1;
    int vs, vt;
    VVI V, revV;
    vector<vector<SubsetPaths>> dp;
    vector<vector<SubsetPaths>> dp_rev;

    // if true, then if L >= N-1, number of all paths will be found using graphillion.
    // thi smight be used only if we are not working on graph with SP2s
    bool use_graphillion_for_all_paths = false;

    void kernelize();

    // void kernelize2();
    VBIG solveBySolver2();


    void readGraph(const bool directed = false);
    static pair<VI,VI> findDistances(VVI & V, int beg, VB & was);
    static pair<VI,VI> findDistances(VVI & V, int beg, Mask S);
    static pair<VI,VI> findDistances(VVI & V, int beg);
    static VI findShortestPath(VVI & V, int beg, int end, Mask S);
    static VI findShortestPath(VVI & V, int beg, int end, VB & was);


    ULL backtracking_states_visited = 0;

    VBIG solveBT();

    // template<bool use_caching, bool use_cch, bool use_sp_dp, bool use_state_merging>
    // VBIG solveBTOptions();

    void solveBruteBT( int num, VB & was, int l, bigint & res );

    auto getDPRange( auto & dpa, auto val, auto fun );
    bool existsPath( VVI & V, int a, int b, Mask & S );

    VBIG calculateResult(VVI & V, VVI & revV, auto & dp, auto & dp_rev, int vs, int vt);
    VBIG calculateResultSpecialCase1(VVI & V, VVI & revV, auto & dp, auto & dp_rev, int vs, int vt);
    void createPaths(int max_length, VVI & V, VVI & revV, int vs, int vt, VI nodes, auto & dp);
    vector<SP> createNewEntries2(VVI & V, auto & dp, int length, VI & nodes);
    vector<SP> createNewEntries(int length, int k1, int k2);
    void trimStates(VVI & V, VVI & revV, int vs, int vt, vector<SubsetPaths> & sps);
    void findD1D2L(VVI & V, VVI & revV, int vs, int vt, int & L, int & D1, int & D2);


    static VBIG countShortestPaths(VVI & V, int vs, VB & was);


    VBIG solveSmartMIM() ;
    VBIG solveMIM2() ;
    VBIG solveSmart();

    Config cnf;
};

#endif //SOLVER_H
