//
// Created by sylwe on 04/09/2024.
//

#ifndef SOLVER2MIM_H
#define SOLVER2MIM_H

#include <Stopwatch.h>

#include "Config.h"
#include "DistanceFinder2.h"
#include "Structures.h"

class Solver2MIM {
public:
    Solver2MIM( GraphSP2 _g, int L, int vs, int vt );

    int N; // size of the input graph
    int L; // max length of paths to consider
    int vs, vt;
    GraphSP2 g, revg;
    Config cnf;

    Stopwatch timer; // used to set maximal possible time...

    // void createPaths(int max_length, GraphSP2 & g, GraphSP2 & revg, int vs, int vt, VI nodes, auto & dp);
    // vector<SP2> createNewEntries(GraphSP2 & g, auto & dp, int length, VI & nodes);
    void trimStates(GraphSP2 & g, GraphSP2 & revg, int vs, int vt, vector<SP2> & sps);

    VBIG solveMIM();
    VBIG solveMIM3();

    static HashType getHashForMask(Mask m, int N);

    // uses biconnected-trimming reduction rule in graph underV \ mask for path beg -> vt.
    // [pth] can be any path from beg to vt - providing any sucgh path makes the algorithm faster
    static Mask getNodesAfterBiconnectedReduction(VVI & underV, Mask & mask, int beg, int vt, VI pth = {});
    static Mask getNodesAfterBiconnectedReduction2(VVI & underV, Mask & mask, int beg, int vt, VI pth = {});


    // same as above, but also changes node hashes accordingly...
    static HashType getMergeStateHashBiconnected(VVI & underV, Mask & mask, int beg, int vt);

    // finds and returns a mask containing nodes that can be on a paired path to state [sp], assuming that
    // it has given [mask]. Here [mask] is given separately, as this function can be used to
    // static Mask getPassableNodes(GraphSP2 & g, GraphSP2& revg, Mask mask, int minl, int beg, int vt, int L );
    static Mask getPassableNodes(GraphSP2 & g, GraphSP2& revg, DistanceFinder2 & df_beg, DistanceFinder2 & df_vt, Mask mask, int minl, int beg, int vt, int L );

    static tuple<HashType,HashType> getAllMergeStateHashes(GraphSP2 & g, GraphSP2 &revg, VVI & underV, SP4 & sp, int beg, int vt, int L);

    // checks if graph G \ S, where set S is marked by set bits in [was] is a DAG
    static bool isDAG( GraphSP2 & G, const VB & was );

    pair<Mask,Mask> createReachableNodeMasksVsVt();

};


#endif //SOLVER2MIM_H
