//
// Created by sylwe on 07/08/2024.
//

#ifndef CCH_H
#define CCH_H

#include "Makros.h"
#include "routingkit/include/routingkit/customizable_contraction_hierarchy.h"

using namespace RoutingKit;

class CCH {
public:
    using VU = vector<unsigned>;
    using VVU = vector<VU>;

    CCH(VVI V, VVI W);

    // finds and returns the distance between nodes a and b
    int dst( int a, int b );
    auto dst( auto sources, auto ends );
    VU dst( int s, VU ends );

    bool hasPath(int a, int b);

    // sets all arcs (v,x) and all arcs (x,v) for which mask[x] == false to [infty]
    void excludeNode(int v, VB & mask);

    // sets all arcs (v,x) and all arcs (x,v) for which mask[x] == false to initial weights
    void includeNode(int v, VB & mask);

    // constructs node order and builds cch for it
    void constructNodeOrderAndBuild();

    // equal to n * max_arc_weight, so that no path in the graph with initially provided weights can have that length
    unsigned infty = 0;

private:

    CustomizableContractionHierarchy cch;
    CustomizableContractionHierarchyMetric metric;
    CustomizableContractionHierarchyQuery query;
    CustomizableContractionHierarchyPartialCustomization partial_customization;

    VU tail, head, weight, init_weights, node_order;
    VU tail_rev, head_rev, weight_rev, init_weights_rev;
    VI first_of, first_of_rev;
    int N;

    void createTailsHeadsWeights(VVI V, VVI W);

    VI rev_arc_id;


    void constructNodeOrder();

};

#endif //CCH_H
