//
// Created by sylwe on 12.07.2024.
//

#ifndef KERNELIZER_H
#define KERNELIZER_H

#include "Config.h"
#include "Makros.h"
#include "Structures.h"

class Core {
public:
    VVI V, revV;
    int vs,vt;
    int L;

    friend ostream& operator<<(ostream & str, Core & c);
};

class Core2 {
public:
    GraphSP2 G;
    int vs, vt;
    int L;

    friend ostream& operator<<(ostream & str, Core2 & c);
};


class Kernelizer {
public:
    static constexpr bool kernelizer_debug = false;

    Kernelizer(VVI & _V, VVI & _revV, int _L, int _vs, int _vt);
    VVI V;
    VVI revV;
    int L,vs,vt;
    VB helper;
    VB was;
    int N;

    Core createCoreGraph();

    bool trimLeaves();

    bool removeDispensibleSCCs();

    bool removeDistantNodes();

    bool removeDistantNodesV2Dominators();

    bool removeDistantArcsV2Dominators();

    bool removeDispensibleBiconnectedComponents();

    bool removeDominatedNodes();

    static void removeArc(VVI & V, VVI & revV, int a, int b);
    static void removeArc( VVI & V, int a, int b );
    static void clearNode(VVI & V, VVI & revV, int v);

    static VI findDominators( VVI & V, int vs, int vt );
    static VI findDominators( VVI & V, VB was, int vs, int vt );

    Config cnf;
};

class Kernelizer2 {
public:
    static constexpr bool kernelizer_debug = true;

    Kernelizer2(VVI & _V, int _L, int _vs, int _vt);
    Kernelizer2(GraphSP2 & _G, int _L, int _vs, int _vt);
    GraphSP2 G;
    // VVI V;
    // VVI revV;
    int L,vs,vt;
    VB helper;
    VB was;
    int N;

    class InducedGraphSP2 {
    public:
        GraphSP2 G;
        VI nodes;
        unordered_map<int,int> perm;

        auto & operator[](int pos){return G[pos];}
        static InducedGraphSP2 induce( GraphSP2 & g, VI nodes );
        static InducedGraphSP2 induceByNonisolatedNodes( GraphSP2 & g );
    };

    Core2 createCore2Graph();


    bool contractMaximalPaths( GraphSP2 & G );
    static void removeArc( GraphSP2 & G, int a, int b );
    static void clearNode(GraphSP2 & g, GraphSP2 & revg, int v);

    bool contractComponents(GraphSP2 & g);
    bool dominatedComponents(GraphSP2 & g, int L, int vs, int vt);

    static void contractNode( int v, GraphSP2 & G, VVI & revG, int L );

    bool removeDistantNodesV2Dominators();
    bool removeDistantArcsV2Dominators();
    bool removeDispensibleSCCs(); // uses version from Kernelizer, since this rule does not use weights...
    bool removeDispensibleBiconnectedComponents(); // uses version from Kernelizer, since this rule does not use weights...
    bool removeDominatedNodes(); // uses version from Kernelizer, since this rule does not use weights...

    Config cnf;
};

#endif //KERNELIZER_H
