//
// Created by sylwe on 04/09/2024.
//

#include "Solver2BT.h"

#include <GraphUtils.h>

#include "BiconnectedFinder.h"
// #include "CCH.h"
#include "CollectionOperators.h"
#include "DistanceFinder2.h"
#include "Solver2.h"
#include "Solver2MIM.h"

Solver2BT::Solver2BT(GraphSP2 _g, int _L, int _vs, int _vt) : g(_g), L(_L), vs(_vs), vt(_vt) {
    N = g.size();
    revg = g.reverseGraph();
}

template<bool use_caching>
VBIG Solver2BT::solveBT() {
    constexpr bool debug = false;
    VBIG res_tab(L+1,0);
    VB was(N);
    backtracking_states_visited = 0;
    VI path;
    if constexpr (debug) DEBUG(L);

    if constexpr (!use_caching) {
        function<void(int,Lengths)> dfs = [&]( int num, Lengths curl ) {
            backtracking_states_visited++;

            if( curl.minl() > L ) return;
            if constexpr (debug){ for(int i=0; i<path.size(); i++) clog << "  "; clog << num << " -> " << curl << endl;}

            if( num == vt ) {
                if constexpr (debug) {
                    path.push_back(num);
                    clog << "Found path: " << path << ", curl: " << curl << endl;
                    path.pop_back();
                }

                for( int i=curl.minl(); i <= min( L, curl.maxl() ); i++ ) res_tab[i] += curl[i];
                return;
            }

            was[num] = true;
            if constexpr (debug) path.push_back(num);
            for( auto & arc : g[num] ) {
                int d = arc.v;
                if(!was[d]) {
                    auto newl = Lengths::multiplyLengths( arc.lengths, curl );
                    if constexpr (debug) clog << "trimming newl: " << newl << " to maxL: " << L << endl;
                    newl.trimToMax(L);
                    dfs( d, newl );
                }
            }
            was[num] = false;
            if constexpr (debug) path.pop_back();
        };

        Lengths curl( VBIG(1,1), 0 );
        dfs(vs, curl);

        return res_tab;
    }else {

        constexpr int HS = (1<<10);
        int HSmask = HS-1;
        vector<map<HashType,Lengths>> cache(HS);
        HashType cur_hash = 0;

        ULL cache_hits = 0;

        function<Lengths(int,int)> dfs = [&]( int num, int curl ) {
            backtracking_states_visited++;

            if( curl > L ) return Lengths(VBIG(1,0),0);
            if constexpr (debug){ for(int i=0; i<path.size(); i++) clog << "  "; clog << num << " -> " << curl << endl;}

            if( num == vt ) {
                if constexpr (debug) {
                    path.push_back(num);
                    clog << "Found path: " << path << ", curl: " << curl << endl;
                    path.pop_back();
                }

                return Lengths(VBIG(1,1),0);
            }

            cur_hash ^= node_hashes[num] ^ node_hashes2[num];
            auto it = cache[cur_hash.hash & HSmask].find(cur_hash);
            if( it != cache[cur_hash.hash & HSmask].end()) {
                cache_hits++;
                cur_hash ^= node_hashes[num] ^ node_hashes2[num];
                return it->second;
            }
            cur_hash ^= node_hashes2[num];

            was[num] = true;
            if constexpr (debug) path.push_back(num);
            Lengths res(VBIG(1,0),0);

            for( auto & arc : g[num] ) {
                int d = arc.v;
                if(!was[d]) {
                    auto temp = dfs( d, curl + arc.minl() );
                    temp *= arc.lengths;
                    temp.trimToMax(L);
                    res += temp;
                }
            }
            was[num] = false;
            if constexpr (debug) path.pop_back();

            cache[(cur_hash^node_hashes2[num]).hash & HSmask][cur_hash^node_hashes2[num]] = res;
            cur_hash ^= node_hashes[num];

            return res;
        };

        Lengths res = dfs(vs, 0);
        for( int i=res.minl(); i <= min( L, res.maxl() ); i++ ) res_tab[i] += res[i];

        int cache_size = accumulate(ALL(cache), 0, [](int s, auto & v){return s + v.size();});
        DEBUG(cache_size);
        DEBUG(cache_hits);
        DEBUG(backtracking_states_visited);

        return res_tab;
    }
}
template VBIG Solver2BT::solveBT<true>();
template VBIG Solver2BT::solveBT<false>();

// VBIG Solver2BT::solveBTCCH() {
//     constexpr bool debug = false;
//     VBIG res_tab(L+1,0);
//     VB was(N);
//     backtracking_states_visited = 0;
//     VI path;
//
//     clog << "Entering Solver2::solveBTCCH" << endl;
//
//     if constexpr (debug) {
//         clog << "In Solver2::solveBTCCH()" << endl;
//         DEBUG(L);
//     }
//
//     VVI V(N), W(N);
//     for(int i=0; i<N; i++) {
//         for( auto & arc : revg[i] ) {
//             V[i].emplace_back(arc.v);
//             W[i].emplace_back(arc.minl());
//         }
//     }
//     CCH cch(V,W); // cch for reverse graph
//     cch.constructNodeOrderAndBuild();
//     if constexpr (debug) {
//         clog << "cch created" << endl;
//         DEBUG(V);
//         DEBUG(W);
//     }
//
//     vector<unsigned> ends;
//     const bool use_shortest_paths = cnf.use_bt_dp_for_shortest_paths;
//     const bool use_caching = cnf.use_bt_caching;
//
//     constexpr int HS = (1<<10);
//     int HSmask = HS-1;
//     vector<map<HashType,Lengths>> cache(HS);
//     HashType cur_hash = 0;
//     ULL cache_hits = 0;
//     const auto zero_lengths = Lengths( VBIG(1,0),0 );
//
//
//     function<Lengths(int,int)> dfs = [&]( int num, int curl ) {
//         if constexpr (debug){ for(int i=0; i<path.size(); i++) clog << "  "; clog << num << " -> " << curl << endl;}
//
//         backtracking_states_visited++;
//         if( num == vt ) {
//             if constexpr (debug) {
//                 for(int i=0; i<path.size(); i++) clog << "  ";
//                 path.push_back(num);
//                 clog << "Found path: " << path << ", curl: " << curl << endl;
//                 path.pop_back();
//             }
//             // for( int i=curl.minl(); i <= min( L, curl.maxl() ); i++ ) res_tab[i] += curl[i];
//             // return;
//             return Lengths( VBIG(1,1),0 );
//         }
//
//
//         HashType state_hash = 0;
//         if (use_caching) {
//             cur_hash ^= node_hashes[num];
//             state_hash = cur_hash ^ node_hashes2[num];
//
//             auto it = cache[state_hash.hash & HSmask].find(cur_hash);
//             if( it != cache[state_hash.hash & HSmask].end()) {
//                 cache_hits++;
//                 cur_hash ^= node_hashes[num];
//                 return it->second;
//             }
//         }
//
//         was[num] = true;
//         cch.excludeNode( num, was );
//
//         ends.clear();
//         for( auto & arc : g[num] ) if(!was[arc.v]) ends.push_back(arc.v);
//         vector<unsigned> distances = cch.dst( vt, ends );
//
//         if constexpr (debug) {
//             for(int i=0; i<path.size(); i++) clog << "  ";
//             clog << "Distances from nodes " << ends << " to " << vt << ": " << distances << endl;
//             path.push_back(num);
//         }
//
//
//         auto R = zero_lengths;
//         int ind = 0;
//         for( auto & arc : g[num] ) {
//             int d = arc.v;
//
//             bigint min_dist_to_vt = L+1;
//             if(!was[d]) {
//                 min_dist_to_vt =  curl + arc.minl() + distances[ind];
//                 ind++;
//             }
//
//             if constexpr (debug) {
//                 if(!was[d]) {
//                     for(int i=0; i<path.size(); i++) clog << "  ";
//                     clog << "path: " << path << ", considering neigh: " << d << " with min_dist_to_vt: " << min_dist_to_vt << endl;
//                 }
//             }
//
//             if(!was[d] && min_dist_to_vt <= L ) {
//                 if( use_shortest_paths && min_dist_to_vt == L && distances[ind-1] + 3 < L) {
//                     auto r = countShortestPaths( g, d, was );
//
//                     if constexpr (debug) {
//                         clog << "Counting shortest paths from d: " << d << " to " << vt << ", S:";
//                         for(int i=0; i<N; i++) if(was[i]) clog << i << ","; clog << endl;
//                         while( r.back() == 0 ) r.pop_back();
//                         for(int i=0; i<N; i++) clog << "r[" << i << "]: " << r[i] << endl;
//                     }
//
//                     auto temp = Lengths( VBIG(1,r[vt]), distances[ind-1] );
//                     temp *= arc.lengths;
//                     R += temp;
//                 }else {
//                     auto temp = dfs( d, curl + arc.minl() );
//                     temp *= arc.lengths;
//                     R += temp;
//                 }
//
//                 R.trimToMax(L);
//             }
//         }
//
//         was[num] = false;
//         if constexpr (debug) {
//             for(int i=0; i<path.size(); i++) clog << "  ";
//             clog << "Including node " << num << " to cch" << endl;
//         }
//         cch.includeNode(num, was);
//
//         if (use_caching) {
//             cache[state_hash.hash & HSmask][state_hash] = R;
//             cur_hash ^= node_hashes[num];
//         }
//
//         if constexpr (debug) path.pop_back();
//
//         return R;
//     };
//
//     auto res = dfs(vs, 0);
//     for(int i=res.minl(); i < min(res.maxl()+1, (int)res_tab.size()); i++) res_tab[i] += res[i];
//
//     if (use_caching) {
//         int cache_size = accumulate(ALL(cache), 0, [](int s, auto & v){return s + v.size();});
//         DEBUG(cache_size);
//         DEBUG(cache_hits);
//     }
//     DEBUG(backtracking_states_visited);
//
//     return res_tab;
// }


VBIG Solver2BT::countShortestPaths(GraphSP2 &G, int vs, auto & was) {
    assert(!was[vs]);
    int N = G.size();

    VI dst(N,inf);
    dst[vs] = 0;
    auto cmp = [&](int a, int b){ if(dst[a] != dst[b]) return dst[a] < dst[b]; return a < b; };
    set<int, decltype(cmp)> zb(cmp);
    for(int i=0; i<N; i++) if(!was[i]) zb.insert(i);

    VBIG res(N,0);
    res[vs] = 1;

    while( !zb.empty() ) {
        int v = *zb.begin();
        zb.erase(zb.begin());

        for( auto & arc : G[v] ) {
            int d = arc.v;
            int l = arc.lengths.minl();
            if(was[d]) continue;

            if( dst[v] + l < dst[d] ) {
                zb.erase(d);
                dst[d] = dst[v] + l;
                zb.insert(d);

                res[d] = 0;
            }

            if( dst[v] + l == dst[d] ) res[d] += res[v] * arc[l];
        }
    }

    return res;
}




VBIG Solver2BT::solveBTStateMerging() {
    constexpr bool debug = true;

    timer.start("bt");
    if( !timer.hasSetLimit("bt") ) timer.setLimit("bt",1e9);

    clog << "Solving using solveBTStateMerging" << endl;

    // if constexpr (debug) { ENDL(2); DEBUG(N); DEBUG(g); DEBUG(L); DEBUG(vs); DEBUG(vt); ENDL(2); }

    Mask was(N);
    backtracking_states_visited = 0;

    vector< unordered_map< HashType, pair<int,Lengths>, HashTypeHash > > cache(N);
    vector< unordered_map< HashType, pair<int,Lengths>, HashTypeHash > > cache2(N);
    HashType cur_hash = 0;
    ULL cache_hits = 0;
    ULL cache_checks = 0;

    ULL shortest_paths_counted = 0;
    ULL zeros_returned = 0;

    auto getCacheEntry = [&]( int v, HashType hash, int l ) -> pair<int,Lengths*> {
        cache_checks++;
        auto it = cache[v].find(hash);
        if( it == cache[v].end() ) return {inf,nullptr};
        cache_hits++;
        return {it->second.first, &it->second.second};
    };

    auto getCacheEntry2 = [&]( int v, HashType hash, int l ) -> pair<int,Lengths*> {
        cache_checks++;
        auto it = cache2[v].find(hash);
        if( it == cache2[v].end() ) return {inf,nullptr};
        cache_hits++;
        return {it->second.first, &it->second.second};
    };

    auto addCacheEntry = [&](int v, HashType hash, int l, Lengths & len) -> Lengths* {
        auto it = cache[v].find(hash);
        if( it == cache[v].end() ) {
            cache[v][hash] = {l,len};
            return &cache[v][hash].second;
        }else if( l < it->second.first ) {
            it->second.second = len;
        }
        return &it->second.second;
    };

    auto addCacheEntry2 = [&](int v, HashType hash, int l, Lengths & len) -> Lengths* {
        auto it = cache2[v].find(hash);
        if( it == cache2[v].end() ) {
            cache2[v][hash] = {l,len};
            return &cache2[v][hash].second;
        }else if( l < it->second.first ) {
            it->second.second = len;
        }
        return &it->second.second;
    };


    VVI underV = g.getStruture();
    VVI V = underV;
    VVI revV = GraphUtils::reverseGraph(underV);
    for( int i=0; i<N; i++ ) {
        underV[i] += revV[i];
        StandardUtils::makeUnique(underV[i]);
    }

    auto zero_lengths = Lengths( VBIG(1,0),0 );
    auto mul_one = Lengths( VBIG(1,1),0 );

    bool result_calculated = true;

    DistanceFinder2 df(N), df2(N);
    BiconnectedFinder bf(N);

    function<Lengths*(int,int)> dfs = [&]( int num, int l ) -> Lengths* {
        backtracking_states_visited++;

        // if( l > L ) return &zero_lengths;
        // if(num == vt) return &mul_one;

        if constexpr (debug) if( backtracking_states_visited % 50'000 == 0 ) {
            int cache_size = accumulate(ALL(cache), 0, [](int s, auto & v){return s + v.size();});
            clog << "\rvisited(k): " << backtracking_states_visited / 1000 << ", checks(k): " << cache_checks / 1000
                << ", hits(k): " << cache_hits/1000 << " (" << (100.0 * cache_hits / cache_checks) << "%)"
                << ", s-paths: " << shortest_paths_counted << ", c-size(k): " << cache_size/1000
                << ", time: " << (int)(timer.getUnstoppedTimeSinceLastStart("bt") / 1000) <<  flush;
        }

        if( timer.tle("bt") ) {
            result_calculated = false;
            return nullptr;
        }

        // auto [dst,par] = Solver2::findDistances(revg,vt,was);
        // if( l + dst[num] > L ) return &zero_lengths;
        df.findDistances( revg, vt,was );
        if( l + df.getDst(num) > L ) {
            zeros_returned++;
            return &zero_lengths;
        }

        VI path;
        // for( int u=num; u != -1; u = par[u] ) path.push_back(u);
        for( int u=num; u != -1; u = df.getPar(u) ) path.push_back(u);
        // Mask biconnected_trimmed = Solver2MIM::getNodesAfterBiconnectedReduction2( underV, was, num, vt, path );
        Mask biconnected_trimmed = bf.getNodesAfterBiconnectedReduction2( underV, was, num, vt, path );
        // VI().swap(par);

        HashType state_bi_hash = Solver2MIM::getHashForMask( biconnected_trimmed,N );
        {
            auto [cl, clen] = getCacheEntry2(num,state_bi_hash,l);
            if(l >= cl) return clen;
        }

        constexpr bool use_distance_trimming = true;
        Mask dst_trimmed = biconnected_trimmed;
        if constexpr (use_distance_trimming){
            // auto dst_num = Solver2::findDistances(g, num, was).first;
            df2.findDistances( g, num, was );

            for( int i=0; i<N; i++ ) {
                // if( dst_trimmed[i] && l + dst_num[i] + dst[i] > L ) dst_trimmed.set(i,false);
                if( dst_trimmed[i] && l + df2.getDst(i) + df.getDst(i) > L ) dst_trimmed.set(i,false);
            }
        }

        HashType state_dst_hash = Solver2MIM::getHashForMask( dst_trimmed,N );
        state_dst_hash ^= ( node_hashes[num] + node_hashes2[num] );
        {
            auto [cl, clen] = getCacheEntry(num,state_dst_hash,l);
            if(l >= cl) return clen;
        }


        // if( l + dst[num] == L ) {
        if( l + df.getDst(num) == L ) {
            shortest_paths_counted++;
            bigint res = countShortestPaths(g, num, was)[vt];
            // Lengths R( dst[num], dst[num] );
            Lengths R( df.getDst(num), df.getDst(num) );
            R.l = VBIG(1,res);
            addCacheEntry2( num, state_bi_hash, l, R );
            return addCacheEntry( num, state_dst_hash, l, R );
        }

        was.set(num,1);

        // sort(ALL(g[num]), [&dst]( auto & a1, auto & a2 ) { return dst[a1.v] < dst[a2.v]; });
        sort(ALL(g[num]), [&df]( auto & a1, auto & a2 ) { return df.getDst(a1.v) < df.getDst(a2.v); });
        VI dst2; dst2.reserve(g[num].size());
        // for( auto & arc : g[num] ) dst2.push_back(dst[arc.v]);
        for( auto & arc : g[num] ) dst2.push_back(df.getDst(arc.v));
        // VI().swap(dst);


        Lengths R( VBIG(1,0),0 );
        for( int i=0; i<g[num].size(); i++ ) {
            auto & arc = g[num][i];

            int d = arc.v;
            if( !biconnected_trimmed[d] || !dst_trimmed[d] ) continue;

            if(!was[d] && l + arc.minl() + dst2[i] <= L ) {
                auto res = (d == vt ? &mul_one : dfs(d, l + arc.minl()) );
                if( res == nullptr ) return nullptr;
                if( !res->valid() ) continue;

                auto temp = arc.lengths;
                temp *= (*res);
                temp.trimToMax(L);
                temp.minimize();
                R += temp;
            }
        }
        was.set(num,0);

        R.trimToMax(L);
        R.minimize();

        addCacheEntry2( num, state_bi_hash,  l, R);
        auto * res = addCacheEntry( num, state_dst_hash,  l, R);

        return res;
    };


    auto lng = dfs(vs,0);
    if( lng == nullptr ) assert( !result_calculated );

    if( !result_calculated ) return {};
    auto lengths = *lng;

    VBIG res(L+1,0);
    for( int i=lengths.minl(); i<min(lengths.maxl()+1, (int)res.size()); i++ ) res[i] += lengths[i];

    int cache_size = accumulate(ALL(cache), 0, [](int s, auto & v){return s + v.size();});
    ENDL(2);
    DEBUG(backtracking_states_visited);
    DEBUG(zeros_returned);
    DEBUG(cache_size);
    DEBUG(cache_hits);
    DEBUG(shortest_paths_counted);
    ENDL(1);


    timer.stop("bt");
    timer.writeAll();

    return res;
}
