//
// Created by sylwe on 04/09/2024.
//

#ifndef SOLVER2BT_H
#define SOLVER2BT_H


#include <Stopwatch.h>

#include "Config.h"
#include "Structures.h"

class Solver2BT {
public:
    Solver2BT( GraphSP2 _g, int L, int vs, int vt );

    int N; // size of the input graph
    int L; // max length of paths to consider
    int vs, vt;
    GraphSP2 g, revg;
    Config cnf;
    ULL backtracking_states_visited = 0;

    Stopwatch timer;

    template<bool use_caching>
    VBIG solveBT();


    VBIG solveBTStateMerging();

    // solves using backtracking and CCH to trim states whenever the distance exceeds L.
    // VBIG solveBTCCH();


    // counts and returns a vector with the number of shortest paths from vs to every other vertex in G \ S,
    // where set S is marked by [was]
    // this is just the usual Dijkstra, so it works in O(M * logN) time (times the time to compare/multiply
    // bigint elements).
    // vertex vs must not be set in [was]
    static VBIG countShortestPaths(GraphSP2 & G, int vs, auto & was);

    // similar to countShortestPaths(), but counts and returns number of shortest paths and number of path of length
    // one more than length of the shortest path
    static pair<VBIG,VBIG> countShortestPathsDeficit1(GraphSP2 & G, int vs, VB & was);

};

#endif //SOLVER2BT_H
