//
// Created by sylwe on 15/09/2024.
//

#ifndef SOLVER2FBSD_H
#define SOLVER2FBSD_H


#include <Stopwatch.h>
#include "Config.h"
#include "Structures.h"

class Solver2FBSD {
public:

    struct Edge {
        Edge( unsigned aa, unsigned bb, int ee1, int ee2 ) : a(aa), b(bb), e1(ee1), e2(ee2) {}
        unsigned int a,b;
        int e1,e2; // ids of arcs that need to be taken to make the transition
        bool operator<(Edge & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    struct CompressedEdge {
        CompressedEdge( unsigned aa, unsigned bb ) : a(aa), b(bb) {}
        unsigned int a,b;
        VI arcs;
        bool operator<(CompressedEdge & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    struct EdgeH {
        EdgeH( HashType aa, HashType bb, int ee1, int ee2 ) : a(aa), b(bb), e1(ee1), e2(ee2) {}
        HashType a,b;
        int e1,e2; // ids of arcs that need to be taken to make the transition
        bool operator<(const EdgeH & oth ) const { return ( a != oth.a ? a < oth.a : b < oth.b ); }
    };

    Solver2FBSD( GraphSP2 _g, int L, int vs, int vt );

    int N; // size of the input graph
    int L; // max length of paths to consider
    int vs, vt;
    GraphSP2 g, revg;
    Config cnf;

    Stopwatch timer;


    /**
     * In this version we construct the transition graph and then calculate the result for the graph.
     * It may be much more memory-consuming than the previous two version, but should also be much faster
     * due to applied optimization techniques...
     */
    VBIG solveFBS3DForNodeOrder( VI order );



};


#endif //SOLVER2FBSD_H
