//
// Created by sylwe on 09/08/2024.
//

#ifndef CONFIG_H
#define CONFIG_H

#include "Makros.h"

class Config{
public:

    static constexpr bool use_bt_instead_of_mim = false; // #TEST - set this value to true for contest, so that BT is used instead of MIM3

    bool solver_use_bt = false;
    bool solver2_use_bt_merging = true; // #TEST - is it equal to or better than MIM3 ?

    bool use_bt_cch = true;
    bool use_bt_caching = true;

    // if true, then if [use_bt_caching] is set then in each state the hash will be change to the one corresponding to
    // 'reachable' nodes from vt, similarly to how it is done in MIM method.
    bool use_bt_advanced_state_merging = false;
    bool use_bt_dp_for_shortest_paths = true;


    bool use_mim = true; // this has precedence over BT, so if this and bt methods are set to use, mim will be used
    bool use_mim_advanced_state_merging = true;
    bool use_mim_bm2 = false;

    bool use_mim_iep = true; // if true, then exclusion-inclusion principle will be used in MIM method
    // only subsets up to this size will be considered for IEP, larger will be compared using pariwise bitmask checks
    int mim_iep_size_threshold = 7;


    bool use_kernelization = true;

    bool use_branching = true;

    bool use_fbs = true; // if true and use_mim == false, then FBS method will be used
    bool select_algorithm = true; // if true, then automatic selection will be done...
    int small_pathwidth_for_fbs = 16;
    int max_pathwidth_for_fbs = 24;
    int small_graph_size_for_mim = 30; // graphs with less than this number of vertives are considered small
    int small_path_length_for_mim = 17; // graphs with less than this number of vertives are considered small
    // int small_length_excess_for_mim = 8; // if L - dst(vs,vt) is small, then use mim

    int max_graph_size_for_mim = 2'000;


    int MAX_RUN_TIME_MILLIS = 10'000;

    // if true, then all data reduction rules will be used even for undirected graph.
    // this might make it impossible to use FBSD algorithm, since it works only for symmetric GraphSP2 objects
    bool use_all_reduction_rules_even_for_undirected_graphs = false;

    int threads = 1; // number of threads to be used

    // creates Json string from this config
    string toString();

};

#endif //CONFIG_H
