//
// Created by sylwe on 12.07.2024.
//

#ifndef STRUCTURES_H
#define STRUCTURES_H

#include <StandardUtils.h>

#include "Makros.h"
#include "Structures.h"

class Bitset {
public:
    static constexpr v8si ones{ (unsigned)-1,(unsigned)-1,(unsigned)-1,(unsigned)-1,
        (unsigned)-1,(unsigned)-1,(unsigned)-1,(unsigned)-1 };
    static constexpr v8si zeros{ 0,0,0,0,0,0,0,0 };

    Bitset( int n, bool val ) { m = vector<v8si>((n+255)/256, val ? ones : zeros); }
    vector<v8si> m;


    bool operator[](idt pos) {
        return (m[pos>>8][(pos&255)>>5] & (1<<(pos&31))) != 0;
    }

    void set(const idt pos, const bool val) {
        if(val) m[pos>>8][(pos&255)>>5] |= (1ll<<(pos&31));
        else m[pos>>8][(pos&255)>>5] &= ~(1<<(pos&31));
    }

    // Bitset operator~(){ Bitset res(*this); for(auto & v : res.m) v = ~v; return res; }
    Bitset& operator&=(const Bitset& oth) {
        // assert(m.size() == oth.m.size());
        for( int i=0; i<m.size(); i++ ) m[i] &= oth.m[i];
        return *this;
    }

    Bitset& operator|=(const Bitset& oth) {
        for( int i=0; i<m.size(); i++ ) m[i] |= oth.m[i];
        return *this;
    }


    int popcnt() {
       int res = 0;
        for(auto & v : m) {
            #pragma gcc unroll 8
            for( int i=0; i<8; i++ ) res += popcount(v[i]);
        }
        return res;
    }
};


class Mask1 {
public:

    // static const BS zeros;
    // static const BS ones;

    Mask1(int N){}
    Mask1( BS mm ) : m(mm) {}
    BS m;

    int size(){ return m.size(); }

    static Mask1 zeros(int N){ return {BS()}; }
    static Mask1 ones(int N){ return {BS()}; }

    bool operator[](idt pos) { return m[pos]; }

    void set(const idt pos, const bool val) { m[pos] = val; }

    Mask1 operator~(){ return ~m; }
    Mask1 operator|(Mask1 & oth){ return m | oth.m; }
    Mask1& operator|=(Mask1 & oth){ m |= oth.m; return *this; }
    Mask1 operator&(Mask1 & oth){ return m & oth.m; }
    Mask1& operator&=(Mask1 & oth){ m &= oth.m; return *this; }
    bool operator==(const Mask1 &oth) const { return m == oth.m; }

    bool operator<(const Mask1 & oth) {
        for( int i=0; i<MAXN; i++ ) if( m[i] != oth.m[i] ) return oth.m[i];
        return false;
    }

    bool singleNodeIntersection( Mask1 & oth ) { return (m & oth.m).count() == 1; }

    bool noIntersection( Mask1 & oth ) { return (m & oth.m).count() == 0; }


    string toString(int maxn = MAXN) {
        string s;
        for(int i=0; i<8; i++) {
            for( int j=0; j<32; j++ ) {
                s += ( (m[i] & (1<<j)) ? "1" : "0" );
                if( s.size() == maxn ) return s;
            }
        }
        return s;
    }

    VI getSetBits() {
        VI res;
        for(int i=0; i<m.size(); i++) if(m[i]) res.push_back(i);
        return res;
    }

    idt popcnt() { return m.count(); }
};

struct Mask2 {
    // Mask() : size(0) {}

    explicit Mask2(int N, bool val = false) : size_n(N){
        // assert(N > 0); // #TEST
        m.resize((N+31) >> 5);
        for(auto & v : m) v = (val ? -1 : 0);
        if( !m.empty() ) m.back() &= lem();
    }

    unsigned lem() {
        if( size_n & 31 ) return (1 << (size_n&31))-1;
        return -1;
    }

    bool operator<( Mask2 & oth ) {
        // DEBUG(toString());
        // DEBUG(oth.toString());
        // DEBUG(m.size());
        // DEBUG(oth.m.size());
        // assert( m.size() == oth.m.size() ); // #TEST
        for(int i=0; i<m.size(); i++) if(m[i] != oth.m[i]) return m[i] < oth.m[i];
        return true;
    }

    static Mask2 zeros(int N ){ return Mask2(N,false); }
    static Mask2 ones(int N ){ return Mask2(N,true); }

    vector<unsigned> m;
private:
    int size_n;
public:
    int size(){ return size_n; }

    bool operator[](int pos) { return m[pos>>5] & (1<<(pos&31)); }
    void set(const int pos, const bool val) {
        if(val) m[pos>>5] |= (1<<(pos&31));
        else m[pos>>5] &= ~(1<<(pos&31));
    }

    void operator|=(Mask2 & oth){ for( int i=0; i<m.size(); i++ ) m[i] |= oth.m[i];  }
    void operator&=(Mask2 & oth){ for( int i=0; i<m.size(); i++ ) m[i] &= oth.m[i]; }
    bool operator==( Mask2 &oth) { for( int i=0; i<m.size(); i++ ) if(m[i] != oth.m[i]) return false; return true; }

    bool singleNodeIntersection( Mask2 & oth ) {
        int res = 0;
        for( int i=0; i<m.size(); i++ ) res += popcount( m[i] & oth.m[i] );
        return res == 1;
    }

    bool noIntersection( Mask2 & oth ) {
        for( int i=0; i<m.size(); i++ ) if( m[i] & oth.m[i] ) return false;
        return true;
    }

    string toString(int maxn = MAXN) {
        string s;
        for(int d : getSetBits()) s += to_string(d) + ",";
        return s;
    }

    VI getSetBits() {
        VI res;
        for( int i=0; i<m.size(); i++ ) for( int j=0; j<32; j++ ) if( m[i] & (1<<j) ) res.push_back( i*32 + j );
        return res;
    }

    void reset(){ for( auto & v : m ) v = 0; }

    int popcnt() { int res = 0; for(unsigned int i : m) res += popcount(i); return res; }
};

using Mask = Mask2;

class HashType {
public:
    HashType() : hash(0){}
    HashType(ULL h) : hash(h) {}
    ULL hash;
    bool operator<( const HashType & oth ) const{ return hash < oth.hash; }
    bool operator==( const HashType & oth ) const{ return hash == oth.hash; }
    HashType operator+(HashType oth){ return hash + oth.hash; }
    HashType& operator+=(HashType oth){ hash += oth.hash; return *this; }
    HashType& operator*=(HashType oth){ hash *= oth.hash; return *this; }
    HashType operator&(HashType oth){ return hash & oth.hash; }
    HashType operator^(HashType oth) const { return hash ^ oth.hash; }
    HashType& operator^=(HashType oth){ hash ^= oth.hash; return *this; }
    HashType operator<<(int shift){ return hash << shift; }
    HashType operator>>(int shift){ return hash >> shift; }

    friend ostream& operator<<( ostream & str, HashType & h );
};
// struct std::hash<HashType>{  std::size_t operator()(const HashType &x) const { return x.hash; } };
struct HashTypeHash { std::size_t operator()(const HashType &x) const { return x.hash; } };




class SubsetPaths {
public:
    // SubsetPaths(){}
    SubsetPaths( Mask m, int ll, HashType h, idt b, idt e, bigint cnt )
    : mask(m), l(ll), hash(h), beg(b), end(e), counts(cnt) {}
    SubsetPaths( const SubsetPaths & oth ) : mask(oth.mask) {
        // mask = oth.mask;
        counts = oth.counts;
        hash = oth.hash;
        beg = oth.beg; end = oth.end; l = oth.l;
    }
    SubsetPaths( SubsetPaths && oth ) : mask(move(oth.mask)) {
        // swap(mask, oth.mask);
        swap(counts, oth.counts);
        hash = oth.hash;
        beg = oth.beg; end = oth.end; l = oth.l;
    }

    Mask mask;
    bigint counts = 0;
    HashType hash;
    idt beg = 0, end = 0, l = 0;

    int length() const{ return l; }

    static bool sortABLM( SubsetPaths & sp1, SubsetPaths & sp2 ) {
        if( sp1.beg != sp2.beg ) return sp1.beg < sp2.beg;
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        if( sp1.l != sp2.l ) return sp1.l < sp2.l;
        if( sp1.hash != sp2.hash ) return sp1.hash < sp2.hash;

        // for( int i=0; i<8; i++ ) if( sp1.mask.m[i] != sp2.mask.m[i]) return sp1.mask.m[i] < sp2.mask.m[i];
        return sp1.mask < sp2.mask;
        return false;
    }

    static bool sortBLM( SubsetPaths & sp1, SubsetPaths & sp2 ) {
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        if( sp1.l != sp2.l ) return sp1.l < sp2.l;
        return sp1.hash < sp2.hash;
    }

    static bool sortLBM( SubsetPaths & sp1, SubsetPaths & sp2 ) {
        if( sp1.l != sp2.l )  return sp1.l < sp2.l;
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        return sp1.hash < sp2.hash;
    }

    string toString() {
        // static_assert( is_same_v<Mask,Mask2> );
        ostringstream str;
        // str << "(mask: " << mask.toString() << ", subset: ";
        str << "(subset: ";
        for( int i=0; i<mask.size(); i++ ) if( mask[i] ) str << (i>0 ? "," : "") << i ;
        str << ", l: " << int(l) << ", beg: " << int(beg) << ", end: " << int(end)
            << ", counts: " << counts << ")";
        return str.str();
    }

    void operator=( const SubsetPaths oth ) {
        mask = oth.mask;
        hash = oth.hash;
        beg = oth.beg;
        end = oth.end;
        l = oth.l;
        counts = oth.counts;
    }

    bool operator==( SubsetPaths & oth) {
        return mask == oth.mask && hash == oth.hash && beg == oth.beg && end == oth.end;
    }
};
using SP = SubsetPaths;

class SubsetPaths3 {
    public:
    // SubsetPaths3(){}
    SubsetPaths3( Mask m, HashType h, bigint cnt ) : mask(m), hash(h), counts(cnt) {}
    SubsetPaths3( SubsetPaths3 & oth ) : mask(oth.mask){ /*mask = oth.mask;*/ counts = oth.counts; hash = oth.hash; }
    SubsetPaths3( SubsetPaths3 && oth ) : mask(oth.mask){ /*swap(mask, oth.mask);*/ swap(counts, oth.counts); hash = oth.hash; }

    Mask mask; // = Mask::zeros();
    bigint counts = 0;
    HashType hash;

    string toString() {
        ostringstream str;
        // str << "(mask: " << mask.toString() << ", subset: ";
        str << "(subset: ";
        for( int i=0; i<mask.size(); i++ ) if( mask[i] ) str << (i>0 ? "," : "") << i ;
        str << ", counts: " << counts << ")";
        return str.str();
    }

    bool operator<(const SubsetPaths3 & oth) const { return hash < oth.hash; }
    bool operator==(const SubsetPaths3 & oth) const{ return hash == oth.hash; }
    void operator=(SubsetPaths3 &oth) {
        mask = oth.mask;
        counts = oth.counts;
        hash = oth.hash;
    }
    void operator=(SubsetPaths3 && oth) {
        swap(mask, oth.mask);
        swap( counts, oth.counts );
        hash = oth.hash;
    }
};
using SP3 = SubsetPaths3;


class Lengths {
public:
    Lengths() : l(VBIG(1,0)), min_l(0){}
    Lengths(int _minl, int _maxl) : min_l(_minl), l(_maxl - _minl + 1,0) {} // size l+1 so that lengths can be addressed up to len inclusive
    Lengths(VBIG len, int _minl) : min_l(_minl), l(len) {} // size l+1 so that lengths can be addressed up to len inclusive
    VBIG l;
    idt min_l = 1;

    void trimToMax(int L){ while( !l.empty() && maxl() > L ) l.pop_back(); }
    bool empty(){ return l.empty(); }
    int minl(){ return min_l; }
    int maxl(){ return minl() + (int)l.size() - 1; }
    int size(){ return l.size(); }
    bigint& operator[](int pos){ return l.at(pos - min_l); }
    friend ostream& operator<<( ostream& str, Lengths & l );

    Lengths operator*( Lengths & oth ) { return multiplyLengths(*this, oth); }
    Lengths& operator*=( Lengths & oth ) {
        if( oth.maxl() == oth.minl() ) {
            for( auto & r : l ) r *= oth[oth.minl()];
            min_l += oth.min_l;
            return *this;
        }
        else if( maxl() == minl() ) { // #TEST #CAUTION - check if this is correct
            auto single = (*this)[minl()];
            l = oth.l;
            for( auto & r : l ) r *= single;
            min_l += oth.minl();
        }
        else {
            auto res = multiplyLengths(*this, oth);
            swap(l, res.l);
            min_l = res.min_l;
        }
        return *this;
    }


    bigint sum(){ return accumulate(ALL(l), bigint(0)); }

    Lengths operator+(Lengths & oth) { return addLengths(*this, oth); }
    Lengths operator-() { for(auto & a : l) a = -a; return *this; }

    Lengths& operator-=(Lengths & oth) {
        auto temp = -oth;
        auto res = addLengths(*this, temp);
        swap(l, res.l);
        min_l = res.min_l;
        return *this;
    }

    Lengths& operator+=(Lengths & oth) {
        if( minl() <= oth.minl() && maxl() >= oth.maxl() ) {
            for( int i=oth.minl(); i <= oth.maxl(); i++ ) (*this)[i] += oth[i];
        } else {
            auto res = addLengths(*this, oth);
            swap(l, res.l);
            min_l = res.min_l;
        }
        return *this;
    }

    static Lengths multiplyLengths(Lengths &l1, Lengths &l2) {
        Lengths res( l1.minl() + l2.minl(), l1.maxl() + l2.maxl() );
        for( int i=l1.minl(); i<=l1.maxl(); i++ ) for( int j=l2.minl(); j <= l2.maxl(); j++ ) res[i+j] += l1[i] * l2[j];
        return res;
    }

    static Lengths addLengths(Lengths &l1, Lengths &l2) {
        Lengths res( min(l1.minl(), l2.minl()), max(l1.maxl(), l2.maxl()) );
        for( int i=res.minl(); i <= res.maxl(); i++ ) {
            if( i >= l1.minl() && i <= l1.maxl() ) res[i] += l1[i];
            if( i >= l2.minl() && i <= l2.maxl() ) res[i] += l2[i];
        }
        return res;
    }

    // for a VBIG [res_tab] with results for path lengths ranging from 0 to res_tab.size()-1, an object Lengths
    // is created.
    static Lengths createFromVBIG(VBIG & res_tab) {
        int ml = 0, Ml = (int)res_tab.size()-1;
        for( int i=0; i<res_tab.size(); i++ ) if( res_tab[i] != 0 ){ ml = i; break; }
        for( int i=(int)res_tab.size()-1; i>=0; i-- ) if( res_tab[i] != 0 ){ Ml = i; break; }

        Lengths lengths( ml,Ml );
        for( int i=ml; i<=Ml; i++ ) lengths[i] = res_tab[i];
        lengths.minimize(); // #TEST - check if it works!
        return lengths;
    }

    bool valid() { return ranges::any_of( l, [](auto & r){ return r != 0; } ); }

    void minimize() {
        int p = 0;
        while( p < l.size() && l[p] == 0 ) p++;
        rotate(l.begin(), l.begin()+p, l.end());
        min_l += p;
        while( !l.empty() && l.back() == 0 ) l.pop_back();
    }

    bool operator==( Lengths & oth ) {
        minimize();
        oth.minimize();
        if(valid() != oth.valid()) return false;
        if( minl() != oth.minl() || maxl() != oth.maxl()) return false;
        for(int i=minl(); i <= maxl(); i++) if( (*this)[i] != oth[i] ) return false;
        return true;
    }

    static HashType hash( bigint & b ) {
        return (ULL)b;
    }

    HashType hash() {
        HashType res = node_hashes2[min_l];
        for( int i=0; i<l.size(); i++ ) {
            HashType h = hash(l[i]);
            h += 1020; // some value here...
            h *= node_hashes[i];
            h += node_hashes2[i];
            res ^= h;
        }
        return res;
    }

};


template<class BI>
class LengthsT {
public:
    using VBIG = vector<BI>;
    using bigint = BI;

    LengthsT() : l(VBIG(1,0)), min_l(0){}
    LengthsT(int _minl, int _maxl) : min_l(_minl), l(_maxl - _minl + 1,0) {} // size l+1 so that lengths can be addressed up to len inclusive
    LengthsT(VBIG len, int _minl) : min_l(_minl), l(len) {} // size l+1 so that lengths can be addressed up to len inclusive
    VBIG l;
    idt min_l = 1;

    void trimToMax(int L){ while( !l.empty() && maxl() > L ) l.pop_back(); }
    bool empty(){ return l.empty(); }
    int minl(){ return min_l; }
    int maxl(){ return minl() + (int)l.size() - 1; }
    int size(){ return l.size(); }
    bigint& operator[](int pos){ return l.at(pos - min_l); }
    // friend ostream& operator<<( ostream& str, LengthsT<BI>& l );

    LengthsT operator*( Lengths & oth ) { return multiplyLengths(*this, oth); }
    LengthsT& operator*=( LengthsT & oth ) {
        if( oth.maxl() == oth.minl() ) {
            for( auto & r : l ) r *= oth[oth.minl()];
            min_l += oth.min_l;
            return *this;
        }
        else {
            auto res = multiplyLengths(*this, oth);
            swap(l, res.l);
            min_l = res.min_l;
        }
        return *this;
    }


    bigint sum(){ return accumulate(ALL(l), bigint(0)); }

    LengthsT operator+(Lengths & oth) { return addLengths(*this, oth); }
    LengthsT operator-() { for(auto & a : l) a = -a; return *this; }

    LengthsT& operator-=(LengthsT & oth) {
        auto temp = -oth;
        auto res = addLengths(*this, temp);
        swap(l, res.l);
        min_l = res.min_l;
        return *this;
    }

    LengthsT& operator+=(LengthsT & oth) {
        auto res = addLengths(*this, oth);
        swap(l, res.l);
        min_l = res.min_l;
        return *this;
    }

    static LengthsT multiplyLengths(LengthsT &l1, LengthsT &l2) {
        LengthsT res( l1.minl() + l2.minl(), l1.maxl() + l2.maxl() );
        for( int i=l1.minl(); i<=l1.maxl(); i++ ) for( int j=l2.minl(); j <= l2.maxl(); j++ ) res[i+j] += l1[i] * l2[j];
        return res;
    }

    static LengthsT addLengths(LengthsT &l1, LengthsT &l2) {
        LengthsT res( min(l1.minl(), l2.minl()), max(l1.maxl(), l2.maxl()) );
        for( int i=res.minl(); i <= res.maxl(); i++ ) {
            if( i >= l1.minl() && i <= l1.maxl() ) res[i] += l1[i];
            if( i >= l2.minl() && i <= l2.maxl() ) res[i] += l2[i];
        }
        return res;
    }

    // for a VBIG [res_tab] with results for path lengths ranging from 0 to res_tab.size()-1, an object Lengths
    // is created.
    static LengthsT createFromVBIG(VBIG & res_tab) {
        int ml = 0, Ml = (int)res_tab.size()-1;
        for( int i=0; i<res_tab.size(); i++ ) if( res_tab[i] != 0 ){ ml = i; break; }
        for( int i=(int)res_tab.size()-1; i>=0; i-- ) if( res_tab[i] != 0 ){ Ml = i; break; }

        LengthsT lengths( ml,Ml );
        for( int i=ml; i<=Ml; i++ ) lengths[i] = res_tab[i];
        lengths.minimize(); // #TEST - check if it works!
        return lengths;
    }

    bool valid() { return ranges::any_of( l, [](auto & r){ return r != 0; } ); }

    void minimize() {
        int p = 0;
        while( p < l.size() && l[p] == 0 ) p++;
        rotate(l.begin(), l.begin()+p, l.end());
        min_l += p;
        while( !l.empty() && l.back() == 0 ) l.pop_back();
    }

    bool operator==( LengthsT & oth ) {
        minimize();
        oth.minimize();
        if(valid() != oth.valid()) return false;
        if( minl() != oth.minl() || maxl() != oth.maxl()) return false;
        for(int i=minl(); i <= maxl(); i++) if( (*this)[i] != oth[i] ) return false;
        return true;
    }

};

template<class BI>
ostream& operator<<( ostream& str, LengthsT<BI> & l ) {
    str << "( minl: " << l.minl() << ", " << l.l << ")";
    return str;
}

class SubsetPaths4 {
public:

    // SubsetPaths4(){}
    SubsetPaths4(int N) : mask(N) {  }
    SubsetPaths4( Mask m, HashType h, Lengths l )
    : mask(m), hash(h), counts(std::move(l)) {}
    SubsetPaths4( const SubsetPaths4 & oth ) : mask(oth.mask) {
        // mask = oth. mask;
        counts = oth.counts;
        hash = oth.hash;
    }
    SubsetPaths4( SubsetPaths4 && oth ) : mask(oth.mask) {
        // swap(mask, oth.mask);
        swap(counts, oth.counts);
        hash = oth.hash;
    }


    Mask mask;
    Lengths counts;
    HashType hash;

    string toString() {
        ostringstream str;
        // str << "(mask: " << mask.toString() << ", subset: ";
        str << "(subset: ";
        for( int i=0; i<mask.size(); i++ ) if( mask[i] ) str << (i>0 ? "," : "") << i ;
        str << ", counts: " << counts << ")";
        return str.str();
    }

    bool operator<(const SubsetPaths4 & oth) const { return hash < oth.hash; }
    bool operator==(const SubsetPaths4 & oth) const { return hash == oth.hash; }
    void operator=(const SubsetPaths4 & oth) {
        mask = oth.mask;
        counts = oth.counts;
        hash = oth.hash;
    }
    void operator=(SubsetPaths4 && oth) {
        swap( mask, oth.mask );
        swap( counts, oth.counts );
        hash = oth.hash;
    }


    friend ostream& operator<<(ostream& str, SubsetPaths4 & sp4);
};
using SP4 = SubsetPaths4;




class SP2Arc {
public:
    SP2Arc(int _v, int minl, int maxl, int initval) : v(_v), lengths(minl, maxl) {
        for( int i=minl; i<=maxl; i++ ) lengths[i] = initval;
    }
    SP2Arc(int _v, Lengths _lengths) : v(_v), lengths(_lengths) {}

    static Lengths mul_one;

    Lengths lengths;
    idt v; // neighbor
    int size(){ return lengths.size(); }
    bigint& operator[](int pos){ return lengths[pos]; }
    int minl(){ return lengths.minl(); }
    int maxl(){ return lengths.maxl(); }
    void resize(int l){ lengths.l.resize(l); }
    friend ostream& operator<<(ostream& str, SP2Arc & arc);
};


class SubsetPaths2 {
    public:
    // SubsetPaths2(){}
    SubsetPaths2( Mask m, VBIG _lengths, int minl, HashType h, idt b, idt e, bigint cnt )
    : mask(m), lengths(_lengths, minl), hash(h), beg(b), end(e), counts(cnt) {}
    Mask mask;
    Lengths lengths;
    bigint counts = 0;
    HashType hash;
    idt beg = 0, end = 0;


    static bool sortABLM( SubsetPaths2 & sp1, SubsetPaths2 & sp2 ) {
        if( sp1.beg != sp2.beg ) return sp1.beg < sp2.beg;
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        // if( sp1.l != sp2.l ) return sp1.l < sp2.l;
        if( sp1.hash != sp2.hash ) return sp1.hash < sp2.hash;

        // for( int i=0; i<8; i++ ) if( sp1.mask.m[i] != sp2.mask.m[i]) return sp1.mask.m[i] < sp2.mask.m[i];
        return sp1.mask < sp2.mask;
        return false;
    }

    static bool sortBLM( SubsetPaths2 & sp1, SubsetPaths2 & sp2 ) {
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        // if( sp1.l != sp2.l ) return sp1.l < sp2.l;
        return sp1.hash < sp2.hash;
    }

    static bool sortLBM( SubsetPaths2 & sp1, SubsetPaths2 & sp2 ) {
        // if( sp1.l != sp2.l )  return sp1.l < sp2.l;
        if( sp1.end != sp2.end ) return sp1.end < sp2.end;
        return sp1.hash < sp2.hash;
    }

    string toString() {
        ostringstream str;
        // str << "(mask: " << mask.toString() << ", subset: ";
        str << "(subset: ";
        for( int i=0; i<mask.size(); i++ ) if( mask[i] ) str << (i>0 ? "," : "") << i ;
        str << ", beg: " << int(beg) << ", end: " << int(end)
            << ", counts: " << counts << ")";
        str << ", lengths: ";
        for( int i=0; i<lengths.size(); i++) str << "(" << i << ": " << lengths[i] << ") ";
        return str.str();
    }

    bool operator==(SubsetPaths2 & oth) {
        return mask == oth.mask && hash == oth.hash && beg == oth.beg && end == oth.end;
    }
};
using SP2 = SubsetPaths2;

/**
 * Graph with edges contatining VBIGs
 */
class GraphSP2 {
public:
    GraphSP2(){}
    GraphSP2( VVI & V );

    // V[i] is a 'neighborhood list' of node i
    // V[i][j] is the j-th SP2Arc A, that leads to node A.v and has admissible lengths A.lengths
    // those lengths can be accessed by [] operator as well
    vector< vector< SP2Arc > > V;


    VVI getStruture();

    // merges multiple edges with the same neighbors of a given node v into single one, by summing lengths
    void mergeParallellEdges(int v);
    void mergeParallellEdges();

    void createFromVVI(VVI & G);

    bool hasArc(int a, int b) {
        for(auto & arc : V[a]) if( arc.v == b ) return true;
        return false;
    }

    SP2Arc& getArc( int a, int b ) {
        for( auto & arc : V[a] ) if( arc.v == b ) return arc;
        assert(false && "no arc found in getSPeArc");
        return V[a][0];
    }

    /**
     * Finds and returns minl value for arc a->b.
     * Required: arc a->b must exists, otherwise an assertion fails or inf is returned.
     */
    int getArcMinl(int a, int b);

    int size(){ return V.size(); }
    auto& operator[](int pos){ return V[pos]; }

    friend ostream& operator<<( ostream & str, GraphSP2 & g );

    int countArcs();

    GraphSP2 reverseGraph();

    bool isSymmetric();

    bool isValid() {
        for( int i=0; i<V.size(); i++ ) {
            for( auto & arc : V[i] ) {
                if(!arc.lengths.valid()) return false;
            }
        }
        return true;
    }

    void removeInvalidArcs() {
        for( auto & v : V ) {
            for( int i=(int)v.size()-1; i>=0; i--) {
                if( !v[i].lengths.valid() ) {
                    swap(v[i],v.back());
                    v.pop_back();
                }
            }
        }
    }

};


class FBSState {
public:
    FBSState(Lengths _counts, VPII _ends, VI _loose, int _vs_node);

    Lengths counts;
    VPII ends;
    VI loose;
    HashType hash;
    int vs_node = -1;

    bool containsNode(int v) {
        for( auto & [a,b] : ends ) if( a == v || b == v ) return true;
        return vs_node == v || StandardUtils::find(loose,v);
    }

    void removeLooseNode(int v) { StandardUtils::removeAll( loose, v ); }
    void removeEnds( PII e ){ StandardUtils::removeAll(ends, e); }

    bool operator==( const FBSState & oth ) const { return hash == oth.hash; }
    bool operator<( FBSState & oth ) const { return hash < oth.hash; }

    friend ostream& operator<<( ostream & str, FBSState & st );

    void recalculateHash(){ hash = getHash(); }

    bool isCorrectState();

    HashType getHash();
};
using FBSS = FBSState;


class FBSState2 {
public:
    FBSState2(Lengths _counts, VPII _ends, int _vs_node);

    Lengths counts;
    VPII ends;
    VI taken;
    HashType hash;
    int vs_node = -1;

    bool containsNode(int v) {
        // for( int a : taken ) if( a == v ) return true;
        for( auto & [a,b] : ends ) if( a == v || b == v ) return true;
        return vs_node == v;
    }
    void removeEnds( PII e ){ StandardUtils::removeAll(ends, e); }

    bool operator==( const FBSState2 & oth ) const { return hash == oth.hash; }
    bool operator<( FBSState2 & oth ) { if( hash != oth.hash ) return hash < oth.hash; else return counts.minl() < oth.counts.minl(); }

    void removeTaken(int v);

    friend ostream& operator<<( ostream & str, FBSState2 & st );

    void recalculateHash(){ hash = getHash(); }

    bool isCorrectState();

    HashType getHash();
};
using FBSS2 = FBSState2;



class FBSState3 {
public:
    FBSState3(){}
    FBSState3(VPII _ends, int _vs_node, int _minl, HashType bndhash);
    // FBSState3(VPII _ends, int _vs_node, int _minl);

    VPII ends;
    VPII taken_masks; // taken masks togehter with minl for those masks...
    HashType hash;
    // HashType bnd_hash;
    int vs_node = -1;
    // unsigned short vs_node = -1;
    int minl = 0;
    // unsigned short minl = 0;

    bool containsNode(int v) {
        // for( int a : taken ) if( a == v ) return true;
        for( auto & [a,b] : ends ) if( a == v || b == v ) return true;
        return vs_node == v;
    }
    void removeEnds( PII e ){ StandardUtils::removeAll(ends, e); }

    bool operator==( const FBSState3 & oth ) const { return hash == oth.hash; }
    bool operator<( FBSState3 & oth ) { if( hash != oth.hash ) return hash < oth.hash; else return minl < oth.minl; }

    void removeTaken(int v, VI & ind_in_S);

    friend ostream& operator<<( ostream & str, FBSState3 & st );

    void recalculateHash(){ hash = getHash(); }
    void recalculateMinl(){ minl = 1e9; for( auto [m,ml] : taken_masks ) minl = min( minl, ml ); }

    bool isCorrectState(VI & ind_in_S);

    static VI getSetBits(int m) {
        VI res;
        for(int i=0; i<32; i++) if( m & (1<<i) ) res.push_back(i);
        return res;
    }

    /**
     * Merges unique masks into single one. If there are many entries with the same mask and different minl values,
     * then smallest minl value is taken
     */
    static void makeUnique( VPII & tm );

    /**
     * Returns two masks - first one contains first at most max_bits set bits of mask m,
     * second one contains last at most max_bits set bits of mask m.
     */
    static PII getTakenMasks(int m, int max_bits);

    HashType getHash();
};
using FBSS3 = FBSState3;


class FBSStateD {
public:
    FBSStateD(Lengths _counts, VPII _ends, int _vs_node);

    Lengths counts;
    VPII ends;
    VI taken;
    HashType hash;
    int vs_node = -1;

    bool containsNode(int v) {
        // for( int a : taken ) if( a == v ) return true;
        for( auto & [a,b] : ends ) if( a == v || b == v ) return true;
        return vs_node == v;
    }
    void removeEnds( PII e ){ if( e.first > e.second ) swap(e.first, e.second); StandardUtils::removeAll(ends, e); }

    bool operator==( const FBSStateD & oth ) const { return hash == oth.hash; }
    bool operator<( FBSStateD & oth ) { if( hash != oth.hash ) return hash < oth.hash; else return counts.minl() < oth.counts.minl(); }

    void removeTaken(int v);

    friend ostream& operator<<( ostream & str, FBSStateD & st );

    void recalculateHash(){ hash = getHash(); }

    bool isCorrectState();

    HashType getHash();
};
using FBSSD = FBSStateD;




class FBSState3D {
public:
    FBSState3D(){}
    FBSState3D(VPII _ends, int _vs_node, int _minl);

    VPII ends;
    VPII taken_masks; // taken masks togehter with minl for those masks...
    HashType hash;
    int vs_node = -1;
    int minl = 0;

    bool containsNode(int v) {
        for( auto & [a,b] : ends ) if( a == v || b == v ) return true;
        return vs_node == v;
    }
    void removeEnds( PII e ) {
        StandardUtils::removeAll(ends, e);
        StandardUtils::removeAll(ends, {e.second, e.first});
    }

    bool operator==( const FBSState3D & oth ) const { return hash == oth.hash; }
    bool operator<( FBSState3D & oth ) const { if( hash != oth.hash ) return hash < oth.hash; else return minl < oth.minl; }

    friend ostream& operator<<( ostream & str, FBSState3D & st );

    void recalculateHash(){ hash = getHash(); }

    void recalculateMinl(){ minl = 1e9; for( auto [m,ml] : taken_masks ) minl = min( minl, ml ); }

    bool isCorrectState(VI & ind_in_S);

    static VI getSetBits(int m) {
        VI res;
        for(int i=0; i<32; i++) if( m & (1<<i) ) res.push_back(i);
        return res;
    }

    /**
     * Merges unique masks into single one. If there are many entries with the same mask and different minl values,
     * then smallest minl value is taken
     */
    static void makeUnique( VPII & tm );

    /**
     * Returns two masks - first one contains first at most max_bits set bits of mask m,
     * second one contains last at most max_bits set bits of mask m.
     */
    static PII getTakenMasks(int m, int max_bits);

    HashType getHash();
};
using FBSS3D = FBSState3D;



#endif //STRUCTURES_H
